from __future__ import annotations

import logging
import time
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, FrozenSet, List, Optional, Sequence, Set, Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.pretty_print import pformat_big_objects
from dbt_semantic_interfaces.protocols.dimension import Dimension, DimensionType
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    DimensionReference,
    MeasureReference,
    MetricReference,
    SemanticModelReference,
    TimeDimensionReference,
)
from dbt_semantic_interfaces.type_enums import MetricType
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.dataset.dataset import DataSet
from metricflow.errors.errors import UnknownMetricLinkingError
from metricflow.model.semantics.linkable_element_properties import LinkableElementProperties
from metricflow.model.semantics.semantic_model_join_evaluator import SemanticModelJoinEvaluator
from metricflow.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow.protocols.semantics import SemanticModelAccessor
from metricflow.specs.specs import (
    DEFAULT_TIME_GRANULARITY,
    DimensionSpec,
    EntityReference,
    EntitySpec,
    LinkableSpecSet,
    TimeDimensionSpec,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ElementPathKey:
    """A key that can uniquely identify an element and the joins used to realize the element."""

    element_name: str
    entity_links: Tuple[EntityReference, ...]
    time_granularity: Optional[TimeGranularity]


@dataclass(frozen=True)
class LinkableDimension:
    """Describes how a dimension can be realized by joining based on entity links."""

    # The semantic model where this dimension was defined.
    semantic_model_origin: SemanticModelReference
    element_name: str
    entity_links: Tuple[EntityReference, ...]
    join_path: Tuple[SemanticModelJoinPathElement, ...]
    properties: FrozenSet[LinkableElementProperties]
    time_granularity: Optional[TimeGranularity] = None

    @property
    def path_key(self) -> ElementPathKey:  # noqa: D
        return ElementPathKey(
            element_name=self.element_name,
            entity_links=self.entity_links,
            time_granularity=self.time_granularity,
        )

    @property
    def reference(self) -> DimensionReference:  # noqa: D
        return DimensionReference(element_name=self.element_name)


@dataclass(frozen=True)
class LinkableEntity:
    """Describes how an entity can be realized by joining based on entity links."""

    # The semantic model where this entity was defined.
    semantic_model_origin: SemanticModelReference
    element_name: str
    properties: FrozenSet[LinkableElementProperties]
    entity_links: Tuple[EntityReference, ...]
    join_path: Tuple[SemanticModelJoinPathElement, ...]

    @property
    def path_key(self) -> ElementPathKey:  # noqa: D
        return ElementPathKey(element_name=self.element_name, entity_links=self.entity_links, time_granularity=None)


@dataclass(frozen=True)
class LinkableElementSet:
    """Container class for storing all linkable elements for a metric.

    TODO: There are similarities with LinkableSpecSet - consider consolidation.
    """

    # Dictionaries that map the path key to context on the dimension
    #
    # For example:
    # {
    #   "listing__country_latest": (
    #     LinkableDimension(
    #       element_name="country_latest",
    #       entity_links=("listing",),
    #       semantic_model_origin="listings_latest_source",
    #   )
    # }
    path_key_to_linkable_dimensions: Dict[ElementPathKey, Tuple[LinkableDimension, ...]]
    path_key_to_linkable_entities: Dict[ElementPathKey, Tuple[LinkableEntity, ...]]

    @staticmethod
    def merge_by_path_key(linkable_element_sets: Sequence[LinkableElementSet]) -> LinkableElementSet:
        """Combine multiple sets together by the path key.

        If there are elements with the same join key, those elements will be categorized as ambiguous.
        """
        key_to_linkable_dimensions: Dict[ElementPathKey, List[LinkableDimension]] = defaultdict(list)
        key_to_linkable_entities: Dict[ElementPathKey, List[LinkableEntity]] = defaultdict(list)

        for linkable_element_set in linkable_element_sets:
            for path_key, linkable_dimensions in linkable_element_set.path_key_to_linkable_dimensions.items():
                key_to_linkable_dimensions[path_key].extend(linkable_dimensions)
            for path_key, linkable_entities in linkable_element_set.path_key_to_linkable_entities.items():
                key_to_linkable_entities[path_key].extend(linkable_entities)

        # Convert the dictionaries to use tuples instead of lists.
        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                path_key: tuple(dimensions) for path_key, dimensions in key_to_linkable_dimensions.items()
            },
            path_key_to_linkable_entities={
                path_key: tuple(entities) for path_key, entities in key_to_linkable_entities.items()
            },
        )

    @staticmethod
    def intersection_by_path_key(linkable_element_sets: Sequence[LinkableElementSet]) -> LinkableElementSet:
        """Find the intersection of all elements in the sets by path key.

        This is useful to figure out the common dimensions that are possible to query with multiple metrics. You would
        find the LinakbleSpecSet for each metric in the query, then do an intersection of the sets.
        """
        if len(linkable_element_sets) == 0:
            return LinkableElementSet(
                path_key_to_linkable_dimensions={},
                path_key_to_linkable_entities={},
            )

        # Find path keys that are common to all LinkableElementSets.
        common_linkable_dimension_path_keys: Set[ElementPathKey] = set.intersection(
            *[
                set(linkable_element_set.path_key_to_linkable_dimensions.keys())
                for linkable_element_set in linkable_element_sets
            ]
        )

        common_linkable_entity_path_keys: Set[ElementPathKey] = set.intersection(
            *[
                set(linkable_element_set.path_key_to_linkable_entities.keys())
                for linkable_element_set in linkable_element_sets
            ]
        )

        # Create a new LinkableElementSet that only includes items where the path key is common to all sets.
        join_path_to_linkable_dimensions: Dict[ElementPathKey, Set[LinkableDimension]] = defaultdict(set)
        join_path_to_linkable_entities: Dict[ElementPathKey, Set[LinkableEntity]] = defaultdict(set)

        for linkable_element_set in linkable_element_sets:
            for path_key, linkable_dimensions in linkable_element_set.path_key_to_linkable_dimensions.items():
                if path_key in common_linkable_dimension_path_keys:
                    join_path_to_linkable_dimensions[path_key].update(linkable_dimensions)
            for path_key, linkable_entities in linkable_element_set.path_key_to_linkable_entities.items():
                if path_key in common_linkable_entity_path_keys:
                    join_path_to_linkable_entities[path_key].update(linkable_entities)

        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                path_key: tuple(
                    sorted(
                        dimensions,
                        key=lambda linkable_dimension: linkable_dimension.semantic_model_origin.semantic_model_name,
                    )
                )
                for path_key, dimensions in join_path_to_linkable_dimensions.items()
            },
            path_key_to_linkable_entities={
                path_key: tuple(
                    sorted(
                        entities, key=lambda linkable_entity: linkable_entity.semantic_model_origin.semantic_model_name
                    )
                )
                for path_key, entities in join_path_to_linkable_entities.items()
            },
        )

    def filter(
        self,
        with_any_of: FrozenSet[LinkableElementProperties],
        without_any_of: FrozenSet[LinkableElementProperties] = frozenset(),
        without_all_of: FrozenSet[LinkableElementProperties] = frozenset(),
    ) -> LinkableElementSet:
        """Filter elements in the set.

        First, only elements with at least one property in the "with_any_of" set are retained. Then, any elements with
        a property in "without_any_of" set are removed. Lastly, any elements with all properties in without_all_of
        are removed.
        """
        key_to_linkable_dimensions: Dict[ElementPathKey, Tuple[LinkableDimension, ...]] = {}
        key_to_linkable_entities: Dict[ElementPathKey, Tuple[LinkableEntity, ...]] = {}

        for path_key, linkable_dimensions in self.path_key_to_linkable_dimensions.items():
            filtered_linkable_dimensions = tuple(
                linkable_dimension
                for linkable_dimension in linkable_dimensions
                if len(linkable_dimension.properties.intersection(with_any_of)) > 0
                and len(linkable_dimension.properties.intersection(without_any_of)) == 0
                and (
                    len(without_all_of) == 0
                    or linkable_dimension.properties.intersection(without_all_of) != without_all_of
                )
            )
            if len(filtered_linkable_dimensions) > 0:
                key_to_linkable_dimensions[path_key] = filtered_linkable_dimensions

        for path_key, linkable_entities in self.path_key_to_linkable_entities.items():
            filtered_linkable_entities = tuple(
                linkable_entity
                for linkable_entity in linkable_entities
                if len(linkable_entity.properties.intersection(with_any_of)) > 0
                and len(linkable_entity.properties.intersection(without_any_of)) == 0
                and (
                    len(without_all_of) == 0
                    or linkable_entity.properties.intersection(without_all_of) != without_all_of
                )
            )
            if len(filtered_linkable_entities) > 0:
                key_to_linkable_entities[path_key] = filtered_linkable_entities

        return LinkableElementSet(
            path_key_to_linkable_dimensions=key_to_linkable_dimensions,
            path_key_to_linkable_entities=key_to_linkable_entities,
        )

    @property
    def as_spec_set(self) -> LinkableSpecSet:  # noqa: D
        return LinkableSpecSet(
            dimension_specs=tuple(
                DimensionSpec(
                    element_name=path_key.element_name,
                    entity_links=path_key.entity_links,
                )
                for path_key in self.path_key_to_linkable_dimensions.keys()
                if not path_key.time_granularity
            ),
            time_dimension_specs=tuple(
                TimeDimensionSpec(
                    element_name=path_key.element_name,
                    entity_links=path_key.entity_links,
                    time_granularity=path_key.time_granularity,
                )
                for path_key in self.path_key_to_linkable_dimensions.keys()
                if path_key.time_granularity
            ),
            entity_specs=tuple(
                EntitySpec(
                    element_name=path_key.element_name,
                    entity_links=path_key.entity_links,
                )
                for path_key in self.path_key_to_linkable_entities
            ),
        )

    @property
    def only_unique_path_keys(self) -> LinkableElementSet:
        """Returns a set that only includes path keys that map to a single element."""
        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                path_key: linkable_dimensions
                for path_key, linkable_dimensions in self.path_key_to_linkable_dimensions.items()
                if len(linkable_dimensions) <= 1
            },
            path_key_to_linkable_entities={
                path_key: linkable_entities
                for path_key, linkable_entities in self.path_key_to_linkable_entities.items()
                if len(linkable_entities) <= 1
            },
        )


@dataclass(frozen=True)
class SemanticModelJoinPathElement:
    """Describes joining a semantic model by the given entity."""

    semantic_model_reference: SemanticModelReference
    join_on_entity: EntityReference


def _generate_linkable_time_dimensions(
    semantic_model_origin: SemanticModelReference,
    dimension: Dimension,
    entity_links: Tuple[EntityReference, ...],
    join_path: Sequence[SemanticModelJoinPathElement],
    with_properties: FrozenSet[LinkableElementProperties],
) -> Sequence[LinkableDimension]:
    """Generates different versions of the given dimension, but at other valid time granularities."""
    linkable_dimensions = []

    defined_time_granularity = (
        dimension.type_params.time_granularity if dimension.type_params else DEFAULT_TIME_GRANULARITY
    )
    for time_granularity in TimeGranularity:
        if time_granularity < defined_time_granularity:
            continue
        properties = set(with_properties)
        if time_granularity != defined_time_granularity:
            properties.add(LinkableElementProperties.DERIVED_TIME_GRANULARITY)

        linkable_dimensions.append(
            LinkableDimension(
                semantic_model_origin=semantic_model_origin,
                element_name=dimension.reference.element_name,
                entity_links=entity_links,
                join_path=tuple(join_path),
                time_granularity=time_granularity,
                properties=frozenset(properties),
            )
        )

    return linkable_dimensions


@dataclass(frozen=True)
class SemanticModelJoinPath:
    """Describes a series of joins between the measure semantic model, and other semantic models by entity.

    For example:

    (measure_source JOIN dimension_source0 ON entity A) JOIN dimension_source1 ON entity B

    would be represented by 2 path elements [(semantic_model0, A), (dimension_source1, B)]
    """

    path_elements: Tuple[SemanticModelJoinPathElement, ...]

    def create_linkable_element_set(
        self, semantic_model_accessor: SemanticModelAccessor, with_properties: FrozenSet[LinkableElementProperties]
    ) -> LinkableElementSet:
        """Given the current path, generate the respective linkable elements from the last semantic model in the path."""
        entity_links = tuple(x.join_on_entity for x in self.path_elements)

        assert len(self.path_elements) > 0
        semantic_model = semantic_model_accessor.get_by_reference(self.path_elements[-1].semantic_model_reference)
        assert semantic_model

        linkable_dimensions: List[LinkableDimension] = []
        linkable_entities: List[LinkableEntity] = []

        for dimension in semantic_model.dimensions:
            dimension_type = dimension.type
            if dimension_type == DimensionType.CATEGORICAL:
                linkable_dimensions.append(
                    LinkableDimension(
                        semantic_model_origin=semantic_model.reference,
                        element_name=dimension.reference.element_name,
                        entity_links=entity_links,
                        join_path=self.path_elements,
                        properties=with_properties,
                    )
                )
            elif dimension_type == DimensionType.TIME:
                linkable_dimensions.extend(
                    _generate_linkable_time_dimensions(
                        semantic_model_origin=semantic_model.reference,
                        dimension=dimension,
                        entity_links=entity_links,
                        join_path=(),
                        with_properties=with_properties,
                    )
                )
            else:
                raise RuntimeError(f"Unhandled type: {dimension_type}")

        for entity in semantic_model.entities:
            # Avoid creating "booking_id__booking_id"
            if entity.reference != entity_links[-1]:
                linkable_entities.append(
                    LinkableEntity(
                        semantic_model_origin=semantic_model.reference,
                        element_name=entity.reference.element_name,
                        entity_links=entity_links,
                        join_path=self.path_elements,
                        properties=with_properties.union({LinkableElementProperties.ENTITY}),
                    )
                )

        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                linkable_dimension.path_key: (linkable_dimension,) for linkable_dimension in linkable_dimensions
            },
            path_key_to_linkable_entities={
                linkable_entity.path_key: (linkable_entity,) for linkable_entity in linkable_entities
            },
        )

    @property
    def last_semantic_model_reference(self) -> SemanticModelReference:
        """The last semantic model that would be joined in this path."""
        assert len(self.path_elements) > 0
        return self.path_elements[-1].semantic_model_reference


class ValidLinkableSpecResolver:
    """Figures out what linkable specs are valid for a given metric.

    e.g. Can you query the metric "bookings" by "listing__country_latest"?
    """

    def __init__(
        self,
        semantic_manifest: SemanticManifest,
        semantic_model_lookup: SemanticModelAccessor,
        max_entity_links: int,
    ) -> None:
        """Constructor.

        Args:
            semantic_manifest: the model to use.
            semantic_model_lookup: used to look up entities for a semantic model.
            max_entity_links: the maximum number of joins to do when computing valid elements.
        """
        self._semantic_manifest = semantic_manifest
        self._semantic_model_lookup = semantic_model_lookup
        # Sort semantic models by name for consistency in building derived objects.
        self._semantic_models = sorted(self._semantic_manifest.semantic_models, key=lambda x: x.name)
        self._join_evaluator = SemanticModelJoinEvaluator(semantic_model_lookup)

        assert max_entity_links >= 0
        self._max_entity_links = max_entity_links

        # Map measures / entities to semantic models that contain them.
        self._entity_to_semantic_model: Dict[str, List[SemanticModel]] = defaultdict(list)
        self._measure_to_semantic_model: Dict[str, List[SemanticModel]] = defaultdict(list)

        for semantic_model in self._semantic_models:
            for entity in semantic_model.entities:
                self._entity_to_semantic_model[entity.reference.element_name].append(semantic_model)

        self._metric_to_linkable_element_sets: Dict[str, List[LinkableElementSet]] = {}

        start_time = time.time()
        for metric in self._semantic_manifest.metrics:
            linkable_sets_for_measure = []
            for measure in metric.measure_references:
                # Cumulative metrics currently can't be queried by other time granularities.
                if metric.type is MetricType.CUMULATIVE:
                    linkable_sets_for_measure.append(
                        self._get_linkable_element_set_for_measure(measure).filter(
                            with_any_of=LinkableElementProperties.all_properties(),
                            without_all_of=frozenset(
                                {
                                    LinkableElementProperties.METRIC_TIME,
                                    LinkableElementProperties.DERIVED_TIME_GRANULARITY,
                                }
                            ),
                        )
                    )
                elif (
                    metric.type is MetricType.SIMPLE
                    or metric.type is MetricType.DERIVED
                    or metric.type is MetricType.RATIO
                ):
                    linkable_sets_for_measure.append(self._get_linkable_element_set_for_measure(measure))
                else:
                    assert_values_exhausted(metric.type)

            self._metric_to_linkable_element_sets[metric.name] = linkable_sets_for_measure
        logger.info(f"Building the [metric -> valid linkable element] index took: {time.time() - start_time:.2f}s")

    def _get_semantic_model_for_measure(self, measure_reference: MeasureReference) -> SemanticModel:  # noqa: D
        semantic_models_where_measure_was_found = []
        for semantic_model in self._semantic_models:
            if any([x.reference.element_name == measure_reference.element_name for x in semantic_model.measures]):
                semantic_models_where_measure_was_found.append(semantic_model)

        if len(semantic_models_where_measure_was_found) == 0:
            raise ValueError(f"No semantic models were found with {measure_reference} in the model")
        elif len(semantic_models_where_measure_was_found) > 1:
            raise ValueError(
                f"Measure {measure_reference} was found in multiple semantic models:\n"
                f"{pformat_big_objects(semantic_models_where_measure_was_found)}"
            )
        return semantic_models_where_measure_was_found[0]

    @staticmethod
    def _get_elements_in_semantic_model(semantic_model: SemanticModel) -> LinkableElementSet:
        """Gets the elements in the semantic model, without requiring any joins.

        Elements related to metric_time are handled separately in _create_elements_for_metric_time().
        """
        linkable_dimensions = []
        linkable_entities = []
        for entity in semantic_model.entities:
            linkable_entities.append(
                LinkableEntity(
                    semantic_model_origin=semantic_model.reference,
                    element_name=entity.reference.element_name,
                    entity_links=(),
                    join_path=(),
                    properties=frozenset({LinkableElementProperties.LOCAL, LinkableElementProperties.ENTITY}),
                )
            )
            for entity_link in SemanticModelLookup.entity_links_for_local_elements(semantic_model):
                linkable_entities.append(
                    LinkableEntity(
                        semantic_model_origin=semantic_model.reference,
                        element_name=entity.reference.element_name,
                        entity_links=(entity_link,),
                        join_path=(),
                        properties=frozenset({LinkableElementProperties.LOCAL, LinkableElementProperties.ENTITY}),
                    )
                )

        for entity_link in SemanticModelLookup.entity_links_for_local_elements(semantic_model):
            dimension_properties = frozenset({LinkableElementProperties.LOCAL})
            for dimension in semantic_model.dimensions:
                dimension_type = dimension.type
                if dimension_type is DimensionType.CATEGORICAL:
                    linkable_dimensions.append(
                        LinkableDimension(
                            semantic_model_origin=semantic_model.reference,
                            element_name=dimension.reference.element_name,
                            entity_links=(entity_link,),
                            join_path=(),
                            properties=dimension_properties,
                        )
                    )
                elif dimension_type is DimensionType.TIME:
                    linkable_dimensions.extend(
                        _generate_linkable_time_dimensions(
                            semantic_model_origin=semantic_model.reference,
                            dimension=dimension,
                            entity_links=(entity_link,),
                            join_path=(),
                            with_properties=dimension_properties,
                        )
                    )
                else:
                    assert_values_exhausted(dimension_type)

        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                linkable_dimension.path_key: (linkable_dimension,) for linkable_dimension in linkable_dimensions
            },
            path_key_to_linkable_entities={
                linkable_entity.path_key: (linkable_entity,) for linkable_entity in linkable_entities
            },
        )

    def _get_semantic_models_with_joinable_entity(
        self,
        left_semantic_model_reference: SemanticModelReference,
        entity_reference: EntityReference,
    ) -> Sequence[SemanticModel]:
        # May switch to non-cached implementation.
        semantic_models = self._entity_to_semantic_model[entity_reference.element_name]
        valid_semantic_models = []
        for semantic_model in semantic_models:
            if self._join_evaluator.is_valid_semantic_model_join(
                left_semantic_model_reference=left_semantic_model_reference,
                right_semantic_model_reference=semantic_model.reference,
                on_entity_reference=entity_reference,
            ):
                valid_semantic_models.append(semantic_model)
        return valid_semantic_models

    @staticmethod
    def _get_time_granularity_for_dimension(
        semantic_model: SemanticModel, time_dimension_reference: TimeDimensionReference
    ) -> TimeGranularity:
        matching_dimensions = tuple(
            dimension
            for dimension in semantic_model.dimensions
            if dimension.type is DimensionType.TIME and dimension.time_dimension_reference == time_dimension_reference
        )

        if len(matching_dimensions) != 1:
            raise RuntimeError(
                f"Did not find a matching time dimension for {time_dimension_reference} in {semantic_model}"
            )
        time_dimension = matching_dimensions[0]
        type_params = time_dimension.type_params
        assert (
            type_params is not None
        ), f"type_params should have been set for {time_dimension_reference} in {semantic_model}"

        assert (
            type_params.time_granularity is not None
        ), f"time_granularity should have been set for {time_dimension_reference} in {semantic_model}"

        return type_params.time_granularity

    def _get_metric_time_elements(self, measure_reference: MeasureReference) -> LinkableElementSet:
        """Create elements for metric_time for a given measure in a semantic model.

        metric_time is a virtual dimension that is the same as aggregation time dimension for a measure, but with a
        different name. Because it doesn't actually exist in the semantic model, these elements need to be created based
        on what aggregation time dimension was used to define the measure.
        """
        measure_semantic_model = self._get_semantic_model_for_measure(measure_reference)

        measure_agg_time_dimension_reference = measure_semantic_model.checked_agg_time_dimension_for_measure(
            measure_reference=measure_reference
        )

        agg_time_dimension_granularity = ValidLinkableSpecResolver._get_time_granularity_for_dimension(
            semantic_model=measure_semantic_model,
            time_dimension_reference=measure_agg_time_dimension_reference,
        )

        # It's possible to aggregate measures to coarser time granularities (except with cumulative metrics).
        possible_metric_time_granularities = tuple(
            time_granularity
            for time_granularity in TimeGranularity
            if agg_time_dimension_granularity.is_smaller_than_or_equal(time_granularity)
        )

        # For each of the possible time granularities, create a LinkableDimension for each one.
        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                ElementPathKey(
                    element_name=DataSet.metric_time_dimension_name(),
                    entity_links=(),
                    time_granularity=time_granularity,
                ): (
                    LinkableDimension(
                        semantic_model_origin=measure_semantic_model.reference,
                        element_name=DataSet.metric_time_dimension_name(),
                        entity_links=(),
                        join_path=(),
                        # Anything that's not at the base time granularity of the measure's aggregation time dimension
                        # should be considered derived.
                        properties=frozenset({LinkableElementProperties.METRIC_TIME})
                        if time_granularity is agg_time_dimension_granularity
                        else frozenset(
                            {
                                LinkableElementProperties.METRIC_TIME,
                                LinkableElementProperties.DERIVED_TIME_GRANULARITY,
                            }
                        ),
                        time_granularity=time_granularity,
                    ),
                )
                for time_granularity in possible_metric_time_granularities
            },
            path_key_to_linkable_entities={},
        )

    def _get_joined_elements(self, measure_semantic_model: SemanticModel) -> LinkableElementSet:
        """Get the elements that can be generated by joining other models to the given model."""
        # Create single-hop elements
        join_paths = []
        for entity in measure_semantic_model.entities:
            semantic_models = self._get_semantic_models_with_joinable_entity(
                left_semantic_model_reference=measure_semantic_model.reference,
                entity_reference=entity.reference,
            )
            for semantic_model in semantic_models:
                if semantic_model.name == measure_semantic_model.name:
                    continue
                join_paths.append(
                    SemanticModelJoinPath(
                        path_elements=(
                            SemanticModelJoinPathElement(
                                semantic_model_reference=semantic_model.reference,
                                join_on_entity=entity.reference,
                            ),
                        )
                    )
                )
        single_hop_elements = LinkableElementSet.merge_by_path_key(
            [
                join_path.create_linkable_element_set(
                    semantic_model_accessor=self._semantic_model_lookup,
                    with_properties=frozenset({LinkableElementProperties.JOINED}),
                )
                for join_path in join_paths
            ]
        )

        # Create multi-hop elements. At each iteration, we generate the list of valid elements based on the current join
        # path, extend all paths to include the next valid semantic model, then repeat.
        multi_hop_elements = LinkableElementSet(path_key_to_linkable_dimensions={}, path_key_to_linkable_entities={})

        for i in range(self._max_entity_links - 1):
            new_join_paths: List[SemanticModelJoinPath] = []
            for join_path in join_paths:
                new_join_paths.extend(
                    self._find_next_possible_paths(
                        measure_semantic_model=measure_semantic_model, current_join_path=join_path
                    )
                )

            if len(new_join_paths) == 0:
                break

            multi_hop_elements = LinkableElementSet.merge_by_path_key(
                (multi_hop_elements,)
                + tuple(
                    new_join_path.create_linkable_element_set(
                        semantic_model_accessor=self._semantic_model_lookup,
                        with_properties=frozenset(
                            {LinkableElementProperties.JOINED, LinkableElementProperties.MULTI_HOP}
                        ),
                    )
                    for new_join_path in new_join_paths
                )
            )
            join_paths = new_join_paths

        return LinkableElementSet.merge_by_path_key((single_hop_elements, multi_hop_elements))

    def _get_linkable_element_set_for_measure(self, measure_reference: MeasureReference) -> LinkableElementSet:
        """Get the valid linkable elements for the given measure."""
        measure_semantic_model = self._get_semantic_model_for_measure(measure_reference)

        elements_in_semantic_model = self._get_elements_in_semantic_model(measure_semantic_model)
        metric_time_elements = self._get_metric_time_elements(measure_reference)
        joined_elements = self._get_joined_elements(measure_semantic_model)

        return LinkableElementSet.merge_by_path_key(
            (
                elements_in_semantic_model,
                metric_time_elements,
                joined_elements,
            )
        )

    def get_linkable_elements_for_metrics(
        self,
        metric_references: Sequence[MetricReference],
        with_any_of: FrozenSet[LinkableElementProperties],
        without_any_of: FrozenSet[LinkableElementProperties],
    ) -> LinkableElementSet:
        """Gets the valid linkable elements that are common to all requested metrics."""
        linkable_element_sets = []
        for metric_reference in metric_references:
            element_sets = self._metric_to_linkable_element_sets.get(metric_reference.element_name)
            if not element_sets:
                raise UnknownMetricLinkingError(f"Unknown metric: {metric_reference} in element set")

            # Using .only_unique_path_keys to exclude ambiguous elements where there are multiple join paths to get
            # a dimension / entity.
            metric_result = LinkableElementSet.intersection_by_path_key(
                [
                    element_set.only_unique_path_keys.filter(with_any_of=with_any_of, without_any_of=without_any_of)
                    for element_set in element_sets
                ]
            )
            linkable_element_sets.append(metric_result)

        intersection_set = LinkableElementSet.intersection_by_path_key(linkable_element_sets)
        return intersection_set

    def _find_next_possible_paths(
        self, measure_semantic_model: SemanticModel, current_join_path: SemanticModelJoinPath
    ) -> Sequence[SemanticModelJoinPath]:
        """Generate the set of possible paths that are 1 semantic model join longer that the "current_join_path"."""
        last_semantic_model_in_path = self._semantic_model_lookup.get_by_reference(
            current_join_path.last_semantic_model_reference
        )
        assert last_semantic_model_in_path
        new_join_paths = []

        for entity in last_semantic_model_in_path.entities:
            entity_reference = entity.reference

            # Don't create cycles in the join path by joining on the same entity.
            if entity_reference in set(x.join_on_entity for x in current_join_path.path_elements):
                continue

            semantic_models_that_can_be_joined = self._get_semantic_models_with_joinable_entity(
                left_semantic_model_reference=last_semantic_model_in_path.reference,
                entity_reference=entity.reference,
            )
            for semantic_model in semantic_models_that_can_be_joined:
                # Don't create cycles in the join path by repeating a semantic model in the path.
                if semantic_model.name == measure_semantic_model.name or any(
                    tuple(
                        path_element.semantic_model_reference == semantic_model.reference
                        for path_element in current_join_path.path_elements
                    )
                ):
                    continue

                new_join_path = SemanticModelJoinPath(
                    path_elements=current_join_path.path_elements
                    + (
                        SemanticModelJoinPathElement(
                            semantic_model_reference=semantic_model.reference, join_on_entity=entity_reference
                        ),
                    )
                )
                new_join_paths.append(new_join_path)

        return new_join_paths



from __future__ import annotations

import logging

from metricflow.dataflow.builder.costing import DefaultCost, DefaultCostFunction
from metricflow.dataflow.dataflow_plan import (
    AggregateMeasuresNode,
    FilterElementsNode,
    JoinDescription,
    JoinToBaseOutputNode,
)
from metricflow.specs.specs import (
    DimensionSpec,
    EntitySpec,
    InstanceSpecSet,
    LinklessEntitySpec,
    MeasureSpec,
    MetricInputMeasureSpec,
)
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository

logger = logging.getLogger(__name__)


def test_costing(consistent_id_object_repository: ConsistentIdObjectRepository) -> None:  # noqa: D
    bookings_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    listings_node = consistent_id_object_repository.simple_model_read_nodes["listings_latest"]

    bookings_spec = MeasureSpec(
        element_name="bookings",
    )
    bookings_filtered = FilterElementsNode(
        parent_node=bookings_node,
        include_specs=InstanceSpecSet(
            measure_specs=(bookings_spec,),
            entity_specs=(
                EntitySpec(
                    element_name="listing",
                    entity_links=(),
                ),
            ),
        ),
    )

    listings_filtered = FilterElementsNode(
        parent_node=listings_node,
        include_specs=InstanceSpecSet(
            dimension_specs=(
                DimensionSpec(
                    element_name="country_latest",
                    entity_links=(),
                ),
            ),
            entity_specs=(
                EntitySpec(
                    element_name="listing",
                    entity_links=(),
                ),
            ),
        ),
    )

    join_node = JoinToBaseOutputNode(
        left_node=bookings_filtered,
        join_targets=[
            JoinDescription(
                join_node=listings_filtered,
                join_on_entity=LinklessEntitySpec.from_element_name("listing"),
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            )
        ],
    )

    bookings_aggregated = AggregateMeasuresNode(
        parent_node=join_node, metric_input_measure_specs=(MetricInputMeasureSpec(measure_spec=bookings_spec),)
    )

    cost_function = DefaultCostFunction()
    cost = cost_function.calculate_cost(bookings_aggregated)

    assert cost == DefaultCost(num_joins=1, num_aggregations=1)



"""Classes required for defining metric definition object instances (see MdoInstance)."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, List, Tuple, TypeVar

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.references import MetricModelReference, SemanticModelElementReference

from metricflow.aggregation_properties import AggregationState
from metricflow.specs.column_assoc import ColumnAssociation
from metricflow.specs.specs import (
    DimensionSpec,
    EntitySpec,
    InstanceSpec,
    InstanceSpecSet,
    MeasureSpec,
    MetadataSpec,
    MetricSpec,
    TimeDimensionSpec,
)

# Type for the specification used in the instance.
SpecT = TypeVar("SpecT", bound=InstanceSpec)


# Type for the column association used in the instance.
ColumnNameAssociationT = TypeVar("ColumnNameAssociationT", bound=ColumnAssociation)


class MdoInstance(ABC, Generic[SpecT]):
    """An instance of a metric definition object.

    An instance is different from the metric definition object in that it correlates to columns in a data set and can be
    in different states. e.g. a measure instance can be aggregated, or a time dimension can be at a different
    granularity.
    """

    # The columns associated with this instance.
    associated_columns: Tuple[ColumnAssociation, ...]
    # The spec that describes this instance.
    spec: SpecT

    @property
    def associated_column(self) -> ColumnAssociation:
        """Helper for getting the associated column until support for multiple associated columns is added."""
        assert len(self.associated_columns) == 1
        return self.associated_columns[0]


# Instances for the major metric object types


@dataclass(frozen=True)
class SemanticModelElementInstance(SerializableDataclass):  # noqa: D
    # This instance is derived from something defined in a semantic model.
    defined_from: Tuple[SemanticModelElementReference, ...]

    @property
    def origin_semantic_model_reference(self) -> SemanticModelElementReference:
        """Property to grab the element reference pointing to the origin semantic model for this element instance.

        By convention this is the zeroth element in the Tuple. At this time these tuples are always of exactly
        length 1, so the simple assertions here work.

        TODO: make this a required input value, rather than a derived property on these objects
        """
        if len(self.defined_from) != 1:
            raise ValueError(
                f"SemanticModelElementInstances should have exactly one entry in the `defined_from` property, because "
                f"otherwise there is no way to ensure that the first element is always the origin semantic model! Found "
                f"{len(self.defined_from)} elements in this particular instance: {self.defined_from}."
            )

        return self.defined_from[0]


@dataclass(frozen=True)
class MeasureInstance(MdoInstance[MeasureSpec], SemanticModelElementInstance):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: MeasureSpec
    aggregation_state: AggregationState


@dataclass(frozen=True)
class DimensionInstance(MdoInstance[DimensionSpec], SemanticModelElementInstance):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: DimensionSpec


@dataclass(frozen=True)
class TimeDimensionInstance(MdoInstance[TimeDimensionSpec], SemanticModelElementInstance):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: TimeDimensionSpec


@dataclass(frozen=True)
class EntityInstance(MdoInstance[EntitySpec], SemanticModelElementInstance):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: EntitySpec


@dataclass(frozen=True)
class MetricInstance(MdoInstance[MetricSpec], SerializableDataclass):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: MetricSpec
    defined_from: Tuple[MetricModelReference, ...]


@dataclass(frozen=True)
class MetadataInstance(MdoInstance[MetadataSpec], SerializableDataclass):  # noqa: D
    associated_columns: Tuple[ColumnAssociation, ...]
    spec: MetadataSpec


# Output type of transform function
TransformOutputT = TypeVar("TransformOutputT")


class InstanceSetTransform(Generic[TransformOutputT], ABC):
    """Function to use for transforming instance sets.

    This function interface should be used instead of manually modifying the members of an instance set so that it's
    easy to track, through "Find Usages", all transformations required for instance sets. This should make adding new
    fields (e.g. WeirdDimensionInstance) to instance sets easier to find and reason about.
    """

    @abstractmethod
    def transform(self, instance_set: InstanceSet) -> TransformOutputT:  # noqa: D
        pass


@dataclass(frozen=True)
class InstanceSet(SerializableDataclass):
    """A set that includes all instance types.

    Generally used to help represent that data that is flowing between nodes in the metric dataflow plan.
    """

    measure_instances: Tuple[MeasureInstance, ...] = ()
    dimension_instances: Tuple[DimensionInstance, ...] = ()
    time_dimension_instances: Tuple[TimeDimensionInstance, ...] = ()
    entity_instances: Tuple[EntityInstance, ...] = ()
    metric_instances: Tuple[MetricInstance, ...] = ()
    metadata_instances: Tuple[MetadataInstance, ...] = ()

    def transform(self, transform_function: InstanceSetTransform[TransformOutputT]) -> TransformOutputT:  # noqa: D
        return transform_function.transform(self)

    @staticmethod
    def merge(instance_sets: List[InstanceSet]) -> InstanceSet:
        """Combine all instances from all instances into a single instance set.

        Instances will be de-duped based on their spec.
        """
        measure_instances: List[MeasureInstance] = []
        dimension_instances: List[DimensionInstance] = []
        time_dimension_instances: List[TimeDimensionInstance] = []
        entity_instances: List[EntityInstance] = []
        metric_instances: List[MetricInstance] = []
        metadata_instances: List[MetadataInstance] = []

        for instance_set in instance_sets:
            for measure_instance in instance_set.measure_instances:
                if measure_instance.spec not in {x.spec for x in measure_instances}:
                    measure_instances.append(measure_instance)
            for dimension_instance in instance_set.dimension_instances:
                if dimension_instance.spec not in {x.spec for x in dimension_instances}:
                    dimension_instances.append(dimension_instance)
            for time_dimension_instance in instance_set.time_dimension_instances:
                if time_dimension_instance.spec not in {x.spec for x in time_dimension_instances}:
                    time_dimension_instances.append(time_dimension_instance)
            for entity_instance in instance_set.entity_instances:
                if entity_instance.spec not in {x.spec for x in entity_instances}:
                    entity_instances.append(entity_instance)
            for metric_instance in instance_set.metric_instances:
                if metric_instance.spec not in {x.spec for x in metric_instances}:
                    metric_instances.append(metric_instance)
            for metadata_instance in instance_set.metadata_instances:
                if metadata_instance.spec not in {x.spec for x in metadata_instances}:
                    metadata_instances.append(metadata_instance)

        return InstanceSet(
            measure_instances=tuple(measure_instances),
            dimension_instances=tuple(dimension_instances),
            time_dimension_instances=tuple(time_dimension_instances),
            entity_instances=tuple(entity_instances),
            metric_instances=tuple(metric_instances),
            metadata_instances=tuple(metadata_instances),
        )

    @property
    def spec_set(self) -> InstanceSpecSet:  # noqa: D
        return InstanceSpecSet(
            measure_specs=tuple(x.spec for x in self.measure_instances),
            dimension_specs=tuple(x.spec for x in self.dimension_instances),
            time_dimension_specs=tuple(x.spec for x in self.time_dimension_instances),
            entity_specs=tuple(x.spec for x in self.entity_instances),
            metric_specs=tuple(x.spec for x in self.metric_instances),
            metadata_specs=tuple(x.spec for x in self.metadata_instances),
        )



from __future__ import annotations

import logging

from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilter
from dbt_semantic_interfaces.references import EntityReference
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.specs.column_assoc import ColumnAssociationResolver
from metricflow.specs.specs import (
    DimensionSpec,
    EntitySpec,
    LinkableSpecSet,
    TimeDimensionSpec,
)
from metricflow.specs.where_filter_transform import WhereSpecFactory

logger = logging.getLogger(__name__)


def test_dimension_in_filter(  # noqa: D
    column_association_resolver: ColumnAssociationResolver,
) -> None:
    where_filter = PydanticWhereFilter(where_sql_template="{{ Dimension('listing__country_latest') }} = 'US'")

    where_filter_spec = WhereSpecFactory(
        column_association_resolver=column_association_resolver,
    ).create_from_where_filter(where_filter)

    assert where_filter_spec.where_sql == "listing__country_latest = 'US'"
    assert where_filter_spec.linkable_spec_set == LinkableSpecSet(
        dimension_specs=(
            DimensionSpec(element_name="country_latest", entity_links=(EntityReference(element_name="listing"),)),
        ),
        time_dimension_specs=(),
        entity_specs=(),
    )


def test_dimension_in_filter_with_grain(  # noqa: D
    column_association_resolver: ColumnAssociationResolver,
) -> None:
    where_filter = PydanticWhereFilter(
        where_sql_template="{{ Dimension('listing__country_latest').grain('WEEK') }} = 'US'"
    )

    where_filter_spec = WhereSpecFactory(
        column_association_resolver=column_association_resolver,
    ).create_from_where_filter(where_filter)

    assert where_filter_spec.where_sql == "listing__country_latest__week = 'US'"
    assert where_filter_spec.linkable_spec_set == LinkableSpecSet(
        dimension_specs=(),
        time_dimension_specs=(
            TimeDimensionSpec(
                element_name="country_latest",
                entity_links=(EntityReference(element_name="listing"),),
                time_granularity=TimeGranularity.WEEK,
            ),
        ),
        entity_specs=(),
    )


def test_time_dimension_in_filter(  # noqa: D
    column_association_resolver: ColumnAssociationResolver,
) -> None:
    where_filter = PydanticWhereFilter(
        where_sql_template="{{ TimeDimension('listing__created_at', 'month') }} = '2020-01-01'"
    )

    where_filter_spec = WhereSpecFactory(
        column_association_resolver=column_association_resolver,
    ).create_from_where_filter(where_filter)

    assert where_filter_spec.where_sql == "listing__created_at__month = '2020-01-01'"
    assert where_filter_spec.linkable_spec_set == LinkableSpecSet(
        dimension_specs=(),
        time_dimension_specs=(
            TimeDimensionSpec(
                element_name="created_at",
                entity_links=(EntityReference(element_name="listing"),),
                time_granularity=TimeGranularity.MONTH,
            ),
        ),
        entity_specs=(),
    )


def test_entity_in_filter(  # noqa: D
    column_association_resolver: ColumnAssociationResolver,
) -> None:
    where_filter = PydanticWhereFilter(
        where_sql_template="{{ Entity('user', entity_path=['listing']) }} == 'example_user_id'"
    )

    where_filter_spec = WhereSpecFactory(
        column_association_resolver=column_association_resolver,
    ).create_from_where_filter(where_filter)

    assert where_filter_spec.where_sql == "listing__user == 'example_user_id'"
    assert where_filter_spec.linkable_spec_set == LinkableSpecSet(
        dimension_specs=(),
        time_dimension_specs=(),
        entity_specs=(EntitySpec(element_name="user", entity_links=(EntityReference(element_name="listing"),)),),
    )



from __future__ import annotations

import logging

from _pytest.fixtures import FixtureRequest
from dbt_semantic_interfaces.references import SemanticModelElementReference

from metricflow.aggregation_properties import AggregationState
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.dataflow_plan import JoinDescription, JoinToBaseOutputNode, ReadSqlSourceNode
from metricflow.dataflow.sql_table import SqlTable
from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.instances import (
    InstanceSet,
    MeasureInstance,
)
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.specs.column_assoc import ColumnAssociation, SingleColumnCorrelationKey
from metricflow.specs.specs import (
    LinklessEntitySpec,
    MeasureSpec,
)
from metricflow.sql.sql_exprs import SqlColumnReference, SqlColumnReferenceExpression
from metricflow.sql.sql_plan import SqlSelectColumn, SqlSelectStatementNode, SqlTableFromClauseNode
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.snapshot_utils import assert_spec_set_snapshot_equal

logger = logging.getLogger(__name__)


def test_no_parent_node_data_set(
    simple_semantic_manifest_lookup: SemanticManifestLookup,
    time_spine_source: TimeSpineSource,
) -> None:
    """Tests getting the data set from a single node."""
    resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
    )

    data_set = SqlDataSet(
        instance_set=InstanceSet(
            measure_instances=(
                MeasureInstance(
                    associated_columns=(
                        ColumnAssociation(
                            column_name="bookings", single_column_correlation_key=SingleColumnCorrelationKey()
                        ),
                    ),
                    defined_from=(
                        SemanticModelElementReference(
                            semantic_model_name="fct_bookings_semantic_model", element_name="bookings"
                        ),
                    ),
                    spec=MeasureSpec(
                        element_name="bookings",
                    ),
                    aggregation_state=AggregationState.NON_AGGREGATED,
                ),
            ),
            dimension_instances=(),
            time_dimension_instances=(),
            entity_instances=(),
        ),
        sql_select_node=SqlSelectStatementNode(
            description="test0",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(SqlColumnReference(table_alias="src", column_name="bookings")),
                    column_alias="bookings",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="fct_bookings")),
            from_source_alias="src",
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
        ),
    )

    node = ReadSqlSourceNode(data_set=data_set)

    assert resolver.get_output_data_set(node).instance_set == data_set.instance_set


def test_joined_node_data_set(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
    time_spine_source: TimeSpineSource,
) -> None:
    """Tests getting the data set from a dataflow plan with a join."""
    resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
    )

    # Join "revenue" with "users_latest" to get "user__home_state_latest"
    revenue_node = consistent_id_object_repository.simple_model_read_nodes["revenue"]
    users_node = consistent_id_object_repository.simple_model_read_nodes["users_latest"]
    join_node = JoinToBaseOutputNode(
        left_node=revenue_node,
        join_targets=[
            JoinDescription(
                join_node=users_node,
                join_on_entity=LinklessEntitySpec.from_element_name("user"),
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            )
        ],
    )

    join_node_output_data_set = resolver.get_output_data_set(join_node)

    assert_spec_set_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        set_id="result0",
        spec_set=join_node_output_data_set.instance_set.spec_set,
    )



from __future__ import annotations

import logging
from dataclasses import dataclass

from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.dataflow.sql_table import SqlTable

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class TimeSpineSource:
    """Defines a source table containing all timestamps to use for computing cumulative metrics."""

    schema_name: str
    table_name: str = "mf_time_spine"
    # Name of the column in the table that contains the dates.
    time_column_name: str = "ds"
    # The time granularity of the dates in the spine table.
    time_column_granularity: TimeGranularity = TimeGranularity.DAY

    @property
    def spine_table(self) -> SqlTable:
        """Table containing all dates."""
        return SqlTable(schema_name=self.schema_name, table_name=self.table_name)



from __future__ import annotations

import logging
import typing
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum
from operator import itemgetter
from typing import Any, Dict, Optional, Sequence

from dbt_semantic_interfaces.implementations.base import FrozenBaseModel
from pydantic import Field

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SqlRequestId:
    """Identifies a request (i.e. a call to SqlClient.query() or SqlClient.execute()) to the SQL engine."""

    id_str: str

    def __repr__(self) -> str:  # noqa: D
        return self.id_str


class SqlRequestTagSet(FrozenBaseModel):
    """Set of tags as a Pydantic model for easy serialization."""

    # Using strings to make for cleaner serialized output. Clients should not use this field directly.
    tag_dict: typing.OrderedDict[str, str] = Field(default_factory=OrderedDict)

    @property
    def tags(self) -> Sequence[SqlRequestTag]:  # noqa: D
        return tuple(SqlRequestTag(key, value) for key, value in self.tag_dict.items())

    @staticmethod
    def create_from_dict(tag_dict: Dict[SqlRequestTagKey, str]) -> SqlRequestTagSet:  # noqa: D
        str_tag_dict = {tag_key_enum.value: value for tag_key_enum, value in tag_dict.items()}
        sorted_tuples = tuple(sorted(str_tag_dict.items(), key=itemgetter(0, 1)))
        return SqlRequestTagSet(tag_dict=OrderedDict(sorted_tuples))

    @staticmethod
    def create_from_request_id(request_id: SqlRequestId) -> SqlRequestTagSet:
        """Create a tag set that only includes the tag for the request ID."""
        tag_dict = OrderedDict()
        tag_dict[SqlRequestTagKey.REQUEST_ID_KEY.value] = request_id.id_str
        return SqlRequestTagSet(tag_dict=tag_dict)

    def add_request_id(self, request_id: SqlRequestId) -> SqlRequestTagSet:
        """Adds the request ID tag to this set."""
        tag_dict = OrderedDict()
        tag_dict[SqlRequestTagKey.REQUEST_ID_KEY.value] = request_id.id_str
        return SqlRequestTagSet.combine((self, SqlRequestTagSet(tag_dict=tag_dict)))

    @staticmethod
    def combine(tag_sets: Sequence[SqlRequestTagSet]) -> SqlRequestTagSet:  # noqa: D
        tag_dict: OrderedDict[str, str] = OrderedDict()
        for tag_set in tag_sets:
            for key, value in tag_set.tag_dict.items():
                if key in tag_dict and tag_dict[key] != value:
                    raise RuntimeError(
                        f"Can't combine tag sets due to a conflicting value for key: {key}. Conflicting values are "
                        f"at least: {value} and {tag_dict[key]}"
                    )
                tag_dict[key] = value

        return SqlRequestTagSet(tag_dict=tag_dict)

    @property
    def request_id(self) -> Optional[SqlRequestId]:
        """The value of the request ID tag."""
        tag_value = self.tag_dict.get(SqlRequestTagKey.REQUEST_ID_KEY.value)
        if tag_value:
            return SqlRequestId(tag_value)
        return None

    def is_subset_of(self, tag_set: SqlRequestTagSet) -> bool:  # noqa: D
        return self.tag_dict.items() <= tag_set.tag_dict.items()


@dataclass(frozen=True)
class SqlRequestTag:
    """A key / value that can be used ot label requests to the SQL engine."""

    key: str
    value: str


class SqlRequestTagKey(Enum):
    """Specific tags used by the system."""

    REQUEST_ID_KEY = "MF_REQUEST_ID"


MF_SYSTEM_TAGS_KEY = "MF_SYSTEM_TAGS"
MF_EXTRA_TAGS_KEY = "MF_EXTRA_TAGS"

# Helps to reduce the need too have "ignore type" everywhere.
JsonDict = Dict[str, Any]  # type: ignore [misc]


class SqlJsonTag:
    """Immutable object that represents a JSON object to be used for tagging SQL requests."""

    def __init__(self, json_dict: Optional[JsonDict] = None) -> None:  # noqa: D
        self._json_dict = OrderedDict(json_dict or {})

    @property
    def json_dict(self) -> JsonDict:  # noqa: D
        return OrderedDict(self._json_dict)

    def combine(self, other_tag: SqlJsonTag) -> SqlJsonTag:  # noqa: D
        new_json_dict = OrderedDict(self._json_dict)
        for k, v in other_tag._json_dict.items():
            if k in new_json_dict:
                logger.error(
                    f"Conflict while combining tags. Conflict key: {k} Conflicting values: {v} and {new_json_dict[k]}"
                )
            new_json_dict[k] = v
        return SqlJsonTag(new_json_dict)

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(json_dict={self._json_dict})"



from __future__ import annotations

import logging

from dbt_semantic_interfaces.pretty_print import pformat_big_objects
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.type_enums import TimeGranularity

from metricflow.dataflow.sql_table import SqlTable
from metricflow.model.semantics.metric_lookup import MetricLookup
from metricflow.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.semantics import SemanticModelAccessor

logger = logging.getLogger(__name__)


class SemanticManifestLookup:
    """Adds semantics information to the user configured model."""

    def __init__(self, semantic_manifest: SemanticManifest) -> None:  # noqa: D
        self._semantic_manifest = semantic_manifest
        self._semantic_model_lookup = SemanticModelLookup(semantic_manifest)
        self._metric_lookup = MetricLookup(self._semantic_manifest, self._semantic_model_lookup)

    @property
    def semantic_manifest(self) -> SemanticManifest:  # noqa: D
        return self._semantic_manifest

    @property
    def semantic_model_lookup(self) -> SemanticModelAccessor:  # noqa: D
        return self._semantic_model_lookup

    @property
    def metric_lookup(self) -> MetricLookup:  # noqa: D
        return self._metric_lookup

    @property
    def time_spine_source(self) -> TimeSpineSource:  # noqa: D
        time_spine_table_configurations = self._semantic_manifest.project_configuration.time_spine_table_configurations

        if not (
            len(time_spine_table_configurations) == 1
            and time_spine_table_configurations[0].grain == TimeGranularity.DAY
        ):
            raise NotImplementedError(
                f"Only a single time spine table configuration with {TimeGranularity.DAY} is currently "
                f"supported. Got:\n"
                f"{pformat_big_objects(time_spine_table_configurations)}"
            )

        time_spine_table_configuration = time_spine_table_configurations[0]
        time_spine_table = SqlTable.from_string(time_spine_table_configuration.location)
        return TimeSpineSource(
            schema_name=time_spine_table.schema_name,
            table_name=time_spine_table.table_name,
            time_column_name=time_spine_table_configuration.column_name,
            time_column_granularity=time_spine_table_configuration.grain,
        )



from __future__ import annotations

import datetime

import pytest
from dbt_semantic_interfaces.references import MetricReference
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataset.dataset import DataSet
from metricflow.filters.time_constraint import TimeRangeConstraint
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository
from metricflow.test.time.metric_time_dimension import MTD_SPEC_DAY, MTD_SPEC_MONTH
from metricflow.time.time_granularity_solver import (
    PartialTimeDimensionSpec,
    RequestTimeGranularityException,
    TimeGranularitySolver,
)


@pytest.fixture(scope="session")
def time_granularity_solver(  # noqa: D
    extended_date_semantic_manifest_lookup: SemanticManifestLookup,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_output_resolver: DataflowPlanNodeOutputDataSetResolver,
) -> TimeGranularitySolver:
    return TimeGranularitySolver(
        semantic_manifest_lookup=extended_date_semantic_manifest_lookup,
        read_nodes=list(consistent_id_object_repository.extended_date_model_read_nodes.values()),
        node_output_resolver=node_output_resolver,
    )


def test_validate_day_granuarity_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.DAY)],
    )


def test_validate_month_granuarity_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.MONTH)],
    )


def test_validate_month_granuarity_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings_monthly")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.MONTH)],
    )


def test_validate_month_granuarity_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings"), MetricReference(element_name="bookings_monthly")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.MONTH)],
    )


def test_validate_year_granularity_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    time_granularity_solver.validate_time_granularity(
        metric_references=[MetricReference(element_name="bookings"), MetricReference(element_name="bookings_monthly")],
        time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.YEAR)],
    )


def test_validate_day_granuarity_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    with pytest.raises(RequestTimeGranularityException):
        time_granularity_solver.validate_time_granularity(
            metric_references=[MetricReference(element_name="bookings_monthly")],
            time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.DAY)],
        )


def test_validate_day_granularity_for_day_and_month_metric(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    with pytest.raises(RequestTimeGranularityException):
        time_granularity_solver.validate_time_granularity(
            metric_references=[
                MetricReference(element_name="bookings"),
                MetricReference(element_name="bookings_monthly"),
            ],
            time_dimension_specs=[DataSet.metric_time_dimension_spec(TimeGranularity.DAY)],
        )


PARTIAL_PTD_SPEC = PartialTimeDimensionSpec(element_name=DataSet.metric_time_dimension_name(), entity_links=())


def test_granularity_solution_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_references=[MetricReference(element_name="bookings")], partial_time_dimension_specs=[PARTIAL_PTD_SPEC]
    ) == {
        PARTIAL_PTD_SPEC: MTD_SPEC_DAY,
    }


def test_granularity_solution_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_references=[MetricReference(element_name="bookings_monthly")],
        partial_time_dimension_specs=[PARTIAL_PTD_SPEC],
    ) == {
        PARTIAL_PTD_SPEC: MTD_SPEC_MONTH,
    }


def test_granularity_solution_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_references=[MetricReference(element_name="bookings"), MetricReference(element_name="bookings_monthly")],
        partial_time_dimension_specs=[PARTIAL_PTD_SPEC],
    ) == {PARTIAL_PTD_SPEC: MTD_SPEC_MONTH}


def test_granularity_error_for_cumulative_metric(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    with pytest.raises(RequestTimeGranularityException):
        time_granularity_solver.validate_time_granularity(
            metric_references=[
                MetricReference(element_name="weekly_bookers"),
                MetricReference(element_name="bookings_monthly"),
            ],
            time_dimension_specs=[MTD_SPEC_MONTH],
        )


def test_adjusted_time_constraint(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 1),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.DAY,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 1, 15),
    )

    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 1),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.MONTH,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 1, 31),
    )

    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 15),
            end_time=datetime.datetime(2020, 2, 15),
        ),
        time_granularity=TimeGranularity.MONTH,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 2, 29),
    )


def test_adjusted_time_constraint_for_week_granularity(
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    """Tests the time range adjustment behavior for weekly granularities."""
    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 8),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.WEEK,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 6),
        end_time=datetime.datetime(2020, 1, 19),
    )


def test_adjusted_time_constraint_at_boundaries(
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    """Tests the time range adjustment behavior at the min / max boundaries."""
    assert (
        time_granularity_solver.adjust_time_range_to_granularity(
            time_range_constraint=TimeRangeConstraint(
                start_time=TimeRangeConstraint.all_time().start_time + datetime.timedelta(days=1),
                end_time=TimeRangeConstraint.all_time().end_time,
            ),
            time_granularity=TimeGranularity.WEEK,
        )
        == TimeRangeConstraint.all_time()
    )

    assert (
        time_granularity_solver.adjust_time_range_to_granularity(
            time_range_constraint=TimeRangeConstraint.all_time(),
            time_granularity=TimeGranularity.WEEK,
        )
        == TimeRangeConstraint.all_time()
    )



from __future__ import annotations

import pytest

from metricflow.dataflow.builder.costing import DefaultCostFunction
from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs.column_assoc import ColumnAssociationResolver
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.fixtures.sql_client_fixtures import sql_client  # noqa: F401, F403

"""
Using 'function' scope to make ID generation more deterministic for the dataflow plan builder fixtures..

Using 'session' scope can result in other 'session' scope fixtures causing ID consistency issues.
"""


@pytest.fixture
def column_association_resolver(  # noqa: D
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> ColumnAssociationResolver:
    return DunderColumnAssociationResolver(simple_semantic_manifest_lookup)


@pytest.fixture
def dataflow_plan_builder(  # noqa: D
    simple_semantic_manifest_lookup: SemanticManifestLookup,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> DataflowPlanBuilder:
    return DataflowPlanBuilder(
        source_nodes=consistent_id_object_repository.simple_model_source_nodes,
        read_nodes=list(consistent_id_object_repository.simple_model_read_nodes.values()),
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
        cost_function=DefaultCostFunction(),
    )


@pytest.fixture
def multihop_dataflow_plan_builder(  # noqa: D
    multi_hop_join_semantic_manifest_lookup: SemanticManifestLookup,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    time_spine_source: TimeSpineSource,
) -> DataflowPlanBuilder:
    return DataflowPlanBuilder(
        source_nodes=consistent_id_object_repository.multihop_model_source_nodes,
        read_nodes=list(consistent_id_object_repository.multihop_model_read_nodes.values()),
        semantic_manifest_lookup=multi_hop_join_semantic_manifest_lookup,
        cost_function=DefaultCostFunction(),
    )


@pytest.fixture
def scd_column_association_resolver(  # noqa: D
    scd_semantic_manifest_lookup: SemanticManifestLookup,
) -> ColumnAssociationResolver:
    return DunderColumnAssociationResolver(scd_semantic_manifest_lookup)


@pytest.fixture
def scd_dataflow_plan_builder(  # noqa: D
    scd_semantic_manifest_lookup: SemanticManifestLookup,
    scd_column_association_resolver: ColumnAssociationResolver,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    time_spine_source: TimeSpineSource,
) -> DataflowPlanBuilder:
    return DataflowPlanBuilder(
        source_nodes=consistent_id_object_repository.scd_model_source_nodes,
        read_nodes=list(consistent_id_object_repository.scd_model_read_nodes.values()),
        semantic_manifest_lookup=scd_semantic_manifest_lookup,
        cost_function=DefaultCostFunction(),
        column_association_resolver=scd_column_association_resolver,
    )


@pytest.fixture(scope="session")
def time_spine_source(  # noqa: D
    sql_client: SqlClient, mf_test_session_state: MetricFlowTestSessionState  # noqa: F811
) -> TimeSpineSource:
    return TimeSpineSource(schema_name=mf_test_session_state.mf_source_schema, table_name="mf_time_spine")



"""Script to help generate test snapshots for all supported engines.

Credentials are stored as a JSON string in an environment variable set via a shell command like:

export MF_TEST_ENGINE_CREDENTIALS=$(cat <<EOF
{
    "duck_db": {
        "engine_url": null,
        "engine_password": null
    },
    "redshift": {
        "engine_url": "redshift://...",
        "engine_password": "..."
    },
    "snowflake": {
        "engine_url": "snowflake://...",
        "engine_password": "..."
    },
    "big_query": {
        "engine_url": "bigquery://",
        "engine_password": "..."
    },
    "databricks": {
        "engine_url": "databricks://...",
        "engine_password": "..."
    },
    "postgres": {
        "engine_url": postgres://...",
        "engine_password": "..."
    },
}
EOF
)
"""
from __future__ import annotations

import logging
import os
from dataclasses import dataclass
from typing import Optional, Sequence

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.implementations.base import FrozenBaseModel
from dbt_semantic_interfaces.pretty_print import pformat_big_objects

from metricflow.protocols.sql_client import SqlEngine

logger = logging.getLogger(__name__)


class MetricFlowTestCredentialSet(FrozenBaseModel):  # noqa: D
    engine_url: Optional[str]
    engine_password: Optional[str]


@dataclass(frozen=True)
class MetricFlowTestConfiguration:  # noqa: D
    engine: SqlEngine
    credential_set: MetricFlowTestCredentialSet


class MetricFlowTestCredentialSetForAllEngines(FrozenBaseModel):  # noqa: D
    duck_db: MetricFlowTestCredentialSet
    redshift: MetricFlowTestCredentialSet
    snowflake: MetricFlowTestCredentialSet
    big_query: MetricFlowTestCredentialSet
    databricks: MetricFlowTestCredentialSet
    postgres: MetricFlowTestCredentialSet

    @property
    def as_configurations(self) -> Sequence[MetricFlowTestConfiguration]:  # noqa: D
        return (
            MetricFlowTestConfiguration(
                engine=SqlEngine.DUCKDB,
                credential_set=self.duck_db,
            ),
            MetricFlowTestConfiguration(
                engine=SqlEngine.REDSHIFT,
                credential_set=self.redshift,
            ),
            MetricFlowTestConfiguration(
                engine=SqlEngine.SNOWFLAKE,
                credential_set=self.snowflake,
            ),
            MetricFlowTestConfiguration(
                engine=SqlEngine.BIGQUERY,
                credential_set=self.big_query,
            ),
            MetricFlowTestConfiguration(
                engine=SqlEngine.DATABRICKS,
                credential_set=self.databricks,
            ),
            MetricFlowTestConfiguration(
                engine=SqlEngine.POSTGRES,
                credential_set=self.postgres,
            ),
        )


SNAPSHOT_GENERATING_TEST_FILES = (
    "metricflow/test/dataflow/optimizer/source_scan/test_cm_branch_combiner.py",
    "metricflow/test/sql/optimizer/test_column_pruner.py",
    "metricflow/test/dataset/test_convert_semantic_model.py",
    "metricflow/test/model/test_data_warehouse_tasks.py",
    "metricflow/test/dataflow/builder/test_dataflow_plan_builder.py",
    "metricflow/test/plan_conversion/test_dataflow_to_execution.py",
    "metricflow/test/plan_conversion/test_dataflow_to_sql_plan.py",
    "metricflow/test/sql/test_engine_specific_rendering.py",
    "metricflow/test/plan_conversion/dataflow_to_sql/test_metric_time_dimension_to_sql.py",
    "metricflow/test/integration/test_rendered_query.py",
    "metricflow/test/sql/optimizer/test_rewriting_sub_query_reducer.py",
    "metricflow/test/sql/test_sql_plan_render.py",
    "metricflow/test/integration/test_rendered_query.py",
    "metricflow/test/sql/optimizer/test_sub_query_reducer.py",
    "metricflow/test/dataflow/optimizer/source_scan/test_source_scan_optimizer.py",
    "metricflow/test/sql/test_sql_plan_render.py",
    "metricflow/test/sql/optimizer/test_sub_query_reducer.py",
    "metricflow/test/sql/optimizer/test_table_alias_simplifier.py",
)


def run_command(command: str) -> None:  # noqa: D
    logger.info(f"Running command {command}")
    return_code = os.system(command)
    if return_code != 0:
        raise RuntimeError(f"Error running command: {command}")


def run_tests(test_configuration: MetricFlowTestConfiguration, test_file_paths: Sequence[str]) -> None:  # noqa: D
    combined_paths = " ".join(test_file_paths)
    if test_configuration.credential_set.engine_url is None:
        if "MF_SQL_ENGINE_URL" in os.environ:
            del os.environ["MF_SQL_ENGINE_URL"]
    else:
        os.environ["MF_SQL_ENGINE_URL"] = test_configuration.credential_set.engine_url

    if test_configuration.credential_set.engine_password is None:
        if "MF_SQL_ENGINE_PASSWORD" in os.environ:
            del os.environ["MF_SQL_ENGINE_PASSWORD"]
    else:
        os.environ["MF_SQL_ENGINE_PASSWORD"] = test_configuration.credential_set.engine_password

    if test_configuration.engine is SqlEngine.DUCKDB:
        # Can't use --use-persistent-source-schema with duckdb since it's in memory.
        run_command(f"pytest -x -vv -n 4 --overwrite-snapshots {combined_paths}")
    elif (
        test_configuration.engine is SqlEngine.REDSHIFT
        or test_configuration.engine is SqlEngine.SNOWFLAKE
        or test_configuration.engine is SqlEngine.BIGQUERY
        or test_configuration.engine is SqlEngine.DATABRICKS
        or test_configuration.engine is SqlEngine.POSTGRES
    ):
        engine_name = test_configuration.engine.value.lower()
        os.environ["MF_TEST_ADAPTER_TYPE"] = engine_name
        hatch_env = f"{engine_name}-env"
        use_persistent_source_schema = test_configuration.engine != SqlEngine.POSTGRES
        run_command(
            f"hatch -v run {hatch_env}:pytest -x -vv -n 4 "
            f"--overwrite-snapshots"
            f"{' --use-persistent-source-schema' if use_persistent_source_schema else ''}"
            f" {combined_paths}"
        )
    else:
        assert_values_exhausted(test_configuration.engine)


def run_cli() -> None:  # noqa: D
    # Setup logging.
    dev_format = "%(asctime)s %(levelname)s %(filename)s:%(lineno)d [%(threadName)s] - %(message)s"
    logging.basicConfig(level=logging.INFO, format=dev_format)

    credential_sets_json_str = os.environ.get("MF_TEST_ENGINE_CREDENTIAL_SETS")
    if credential_sets_json_str is None:
        raise ValueError(
            f"Environment variable: MF_TEST_ENGINE_CREDENTIAL_SETS has not been set. Please see the comment in "
            f"{__file__} for details on how to set it."
        )

    credential_sets = MetricFlowTestCredentialSetForAllEngines.parse_raw(credential_sets_json_str)

    logger.info(
        f"Running the following tests to generate snapshots:\n{pformat_big_objects(SNAPSHOT_GENERATING_TEST_FILES)}"
    )

    for test_configuration in credential_sets.as_configurations:
        logger.info(
            f"Running tests for {test_configuration.engine} with URL: {test_configuration.credential_set.engine_url}"
        )
        run_tests(test_configuration, SNAPSHOT_GENERATING_TEST_FILES)


if __name__ == "__main__":
    run_cli()



from __future__ import annotations

import logging
from collections import OrderedDict
from typing import List, Optional, Sequence, Union

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols.metric import MetricType
from dbt_semantic_interfaces.references import MetricModelReference
from dbt_semantic_interfaces.type_enums.aggregation_type import AggregationType

from metricflow.aggregation_properties import AggregationState
from metricflow.dag.id_generation import IdGeneratorRegistry
from metricflow.dataflow.dataflow_plan import (
    AggregateMeasuresNode,
    BaseOutput,
    CombineMetricsNode,
    ComputedMetricsOutput,
    ComputeMetricsNode,
    ConstrainTimeRangeNode,
    DataflowPlanNodeVisitor,
    FilterElementsNode,
    JoinAggregatedMeasuresByGroupByColumnsNode,
    JoinOverTimeRangeNode,
    JoinToBaseOutputNode,
    JoinToTimeSpineNode,
    MetricTimeDimensionTransformNode,
    OrderByLimitNode,
    ReadSqlSourceNode,
    SemiAdditiveJoinNode,
    WhereConstraintNode,
    WriteToResultDataframeNode,
    WriteToResultTableNode,
)
from metricflow.dataset.dataset import DataSet
from metricflow.dataset.sql_dataset import SqlDataSet
from metricflow.filters.time_constraint import TimeRangeConstraint
from metricflow.instances import (
    InstanceSet,
    MetricInstance,
    TimeDimensionInstance,
)
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.instance_converters import (
    AddLinkToLinkableElements,
    AddMetrics,
    AliasAggregatedMeasures,
    ChangeAssociatedColumns,
    ChangeMeasureAggregationState,
    CreateSelectColumnsForInstances,
    CreateSelectColumnsWithMeasuresAggregated,
    FilterElements,
    FilterLinkableInstancesWithLeadingLink,
    RemoveMeasures,
    RemoveMetrics,
    create_select_columns_for_instance_sets,
)
from metricflow.plan_conversion.select_column_gen import (
    SelectColumnSet,
)
from metricflow.plan_conversion.spec_transforms import (
    CreateColumnAssociations,
    CreateSelectCoalescedColumnsForLinkableSpecs,
    SelectOnlyLinkableSpecs,
)
from metricflow.plan_conversion.sql_join_builder import (
    AnnotatedSqlDataSet,
    ColumnEqualityDescription,
    SqlQueryPlanJoinBuilder,
)
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlEngine
from metricflow.specs.column_assoc import ColumnAssociation, ColumnAssociationResolver, SingleColumnCorrelationKey
from metricflow.specs.specs import (
    MeasureSpec,
    MetricSpec,
    TimeDimensionSpec,
)
from metricflow.sql.optimizer.optimization_levels import (
    SqlQueryOptimizationLevel,
    SqlQueryOptimizerConfiguration,
)
from metricflow.sql.sql_exprs import (
    SqlBetweenExpression,
    SqlColumnReference,
    SqlColumnReferenceExpression,
    SqlComparison,
    SqlComparisonExpression,
    SqlDateTruncExpression,
    SqlExpressionNode,
    SqlExtractExpression,
    SqlFunctionExpression,
    SqlLogicalExpression,
    SqlLogicalOperator,
    SqlRatioComputationExpression,
    SqlStringExpression,
    SqlStringLiteralExpression,
)
from metricflow.sql.sql_plan import (
    SqlJoinDescription,
    SqlJoinType,
    SqlOrderByDescription,
    SqlQueryPlan,
    SqlQueryPlanNode,
    SqlSelectColumn,
    SqlSelectStatementNode,
    SqlTableFromClauseNode,
)
from metricflow.time.time_constants import ISO8601_PYTHON_FORMAT

logger = logging.getLogger(__name__)


def _make_time_range_comparison_expr(
    table_alias: str, column_alias: str, time_range_constraint: TimeRangeConstraint
) -> SqlExpressionNode:
    """Build an expression like "ds BETWEEN CAST('2020-01-01' AS TIMESTAMP) AND CAST('2020-01-02' AS TIMESTAMP)."""
    # TODO: Update when adding < day granularity support.
    return SqlBetweenExpression(
        column_arg=SqlColumnReferenceExpression(
            SqlColumnReference(
                table_alias=table_alias,
                column_name=column_alias,
            )
        ),
        start_expr=SqlStringLiteralExpression(
            literal_value=time_range_constraint.start_time.strftime(ISO8601_PYTHON_FORMAT),
        ),
        end_expr=SqlStringLiteralExpression(
            literal_value=time_range_constraint.end_time.strftime(ISO8601_PYTHON_FORMAT),
        ),
    )


class DataflowToSqlQueryPlanConverter(DataflowPlanNodeVisitor[SqlDataSet]):
    """Generates an SQL query plan from a node in the a metric dataflow plan."""

    def __init__(
        self,
        column_association_resolver: ColumnAssociationResolver,
        semantic_manifest_lookup: SemanticManifestLookup,
    ) -> None:
        """Constructor.

        Args:
            column_association_resolver: controls how columns for instances are generated and used between nested
            queries.
            semantic_manifest_lookup: Self-explanatory.
        """
        self._column_association_resolver = column_association_resolver
        self._metric_lookup = semantic_manifest_lookup.metric_lookup
        self._semantic_model_lookup = semantic_manifest_lookup.semantic_model_lookup
        self._time_spine_source = semantic_manifest_lookup.time_spine_source

    @property
    def column_association_resolver(self) -> ColumnAssociationResolver:  # noqa: D
        return self._column_association_resolver

    def _next_unique_table_alias(self) -> str:
        """Return the next unique table alias to use in generating queries."""
        return IdGeneratorRegistry.for_class(self.__class__).create_id(prefix="subq")

    def _make_time_spine_data_set(
        self,
        metric_time_dimension_instance: TimeDimensionInstance,
        metric_time_dimension_column_name: str,
        time_spine_source: TimeSpineSource,
        time_range_constraint: Optional[TimeRangeConstraint] = None,
    ) -> SqlDataSet:
        """Make a time spine data set, which contains all date values like '2020-01-01', '2020-01-02'...

        This is useful in computing cumulative metrics. This will need to be updated to support granularities finer than a
        day.
        """
        time_spine_instance = (
            TimeDimensionInstance(
                defined_from=metric_time_dimension_instance.defined_from,
                associated_columns=(
                    ColumnAssociation(
                        column_name=metric_time_dimension_column_name,
                        single_column_correlation_key=SingleColumnCorrelationKey(),
                    ),
                ),
                spec=metric_time_dimension_instance.spec,
            ),
        )
        time_spine_instance_set = InstanceSet(
            time_dimension_instances=time_spine_instance,
        )
        description = "Date Spine"
        time_spine_table_alias = self._next_unique_table_alias()

        # If the requested granularity is the same as the granularity of the spine, do a direct select.
        if metric_time_dimension_instance.spec.time_granularity == time_spine_source.time_column_granularity:
            return SqlDataSet(
                instance_set=time_spine_instance_set,
                sql_select_node=SqlSelectStatementNode(
                    description=description,
                    # This creates select expressions for all columns referenced in the instance set.
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                SqlColumnReference(
                                    table_alias=time_spine_table_alias,
                                    column_name=time_spine_source.time_column_name,
                                ),
                            ),
                            column_alias=metric_time_dimension_column_name,
                        ),
                    ),
                    from_source=SqlTableFromClauseNode(sql_table=time_spine_source.spine_table),
                    from_source_alias=time_spine_table_alias,
                    joins_descs=(),
                    group_bys=(),
                    where=_make_time_range_comparison_expr(
                        table_alias=time_spine_table_alias,
                        column_alias=time_spine_source.time_column_name,
                        time_range_constraint=time_range_constraint,
                    )
                    if time_range_constraint
                    else None,
                    order_bys=(),
                ),
            )
        # If the granularity is different, apply a DATE_TRUNC() and aggregate.
        else:
            select_columns = (
                SqlSelectColumn(
                    expr=SqlDateTruncExpression(
                        time_granularity=metric_time_dimension_instance.spec.time_granularity,
                        arg=SqlColumnReferenceExpression(
                            SqlColumnReference(
                                table_alias=time_spine_table_alias,
                                column_name=time_spine_source.time_column_name,
                            ),
                        ),
                    ),
                    column_alias=metric_time_dimension_column_name,
                ),
            )
            return SqlDataSet(
                instance_set=time_spine_instance_set,
                sql_select_node=SqlSelectStatementNode(
                    description=description,
                    # This creates select expressions for all columns referenced in the instance set.
                    select_columns=select_columns,
                    from_source=SqlTableFromClauseNode(sql_table=time_spine_source.spine_table),
                    from_source_alias=time_spine_table_alias,
                    joins_descs=(),
                    group_bys=select_columns,
                    where=_make_time_range_comparison_expr(
                        table_alias=time_spine_table_alias,
                        column_alias=time_spine_source.time_column_name,
                        time_range_constraint=time_range_constraint,
                    )
                    if time_range_constraint
                    else None,
                    order_bys=(),
                ),
            )

    def visit_source_node(self, node: ReadSqlSourceNode) -> SqlDataSet:
        """Generate the SQL to read from the source."""
        return SqlDataSet(
            sql_select_node=node.data_set.sql_select_node,
            instance_set=node.data_set.instance_set,
        )

    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode) -> SqlDataSet:
        """Generate time range join SQL."""
        table_alias_to_instance_set: OrderedDict[str, InstanceSet] = OrderedDict()

        input_data_set = node.parent_node.accept(self)
        input_data_set_alias = self._next_unique_table_alias()

        metric_time_dimension_spec: Optional[TimeDimensionSpec] = None
        metric_time_dimension_instance: Optional[TimeDimensionInstance] = None
        for instance in input_data_set.metric_time_dimension_instances:
            if len(instance.spec.entity_links) == 0:
                metric_time_dimension_instance = instance
                metric_time_dimension_spec = instance.spec
                break

        assert metric_time_dimension_spec
        time_spine_data_set_alias = self._next_unique_table_alias()

        metric_time_dimension_column_name = self.column_association_resolver.resolve_spec(
            metric_time_dimension_spec
        ).column_name

        # Assemble time_spine dataset with metric_time_dimension to join.
        # Granularity of time_spine column should match granularity of metric_time column from parent dataset.
        assert metric_time_dimension_instance
        time_spine_data_set = self._make_time_spine_data_set(
            metric_time_dimension_instance=metric_time_dimension_instance,
            metric_time_dimension_column_name=metric_time_dimension_column_name,
            time_spine_source=self._time_spine_source,
            time_range_constraint=node.time_range_constraint,
        )
        table_alias_to_instance_set[time_spine_data_set_alias] = time_spine_data_set.instance_set

        # Figure out which columns correspond to the time dimension that we want to join on.
        input_data_set_metric_time_column_association = input_data_set.column_association_for_time_dimension(
            metric_time_dimension_spec
        )
        input_data_set_metric_time_col = input_data_set_metric_time_column_association.column_name

        time_spine_data_set_column_associations = time_spine_data_set.column_association_for_time_dimension(
            metric_time_dimension_spec
        )
        time_spine_data_set_time_dimension_col = time_spine_data_set_column_associations.column_name

        annotated_input_data_set = AnnotatedSqlDataSet(
            data_set=input_data_set, alias=input_data_set_alias, _metric_time_column_name=input_data_set_metric_time_col
        )
        annotated_time_spine_data_set = AnnotatedSqlDataSet(
            data_set=time_spine_data_set,
            alias=time_spine_data_set_alias,
            _metric_time_column_name=time_spine_data_set_time_dimension_col,
        )

        join_desc = SqlQueryPlanJoinBuilder.make_cumulative_metric_time_range_join_description(
            node=node, metric_data_set=annotated_input_data_set, time_spine_data_set=annotated_time_spine_data_set
        )

        modified_input_instance_set = InstanceSet(
            measure_instances=input_data_set.instance_set.measure_instances,
            dimension_instances=input_data_set.instance_set.dimension_instances,
            entity_instances=input_data_set.instance_set.entity_instances,
            metric_instances=input_data_set.instance_set.metric_instances,
            # we omit the metric time dimension from the right side of the self-join because we need to use
            # the metric time dimension from the right side
            time_dimension_instances=tuple(
                [
                    time_dimension_instance
                    for time_dimension_instance in input_data_set.instance_set.time_dimension_instances
                    if time_dimension_instance.spec != metric_time_dimension_spec
                ]
            ),
        )
        table_alias_to_instance_set[input_data_set_alias] = modified_input_instance_set

        # The output instances are the same as the input instances.
        output_instance_set = ChangeAssociatedColumns(self._column_association_resolver).transform(
            input_data_set.instance_set
        )

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=create_select_columns_for_instance_sets(
                    self._column_association_resolver, table_alias_to_instance_set
                ),
                from_source=time_spine_data_set.sql_select_node,
                from_source_alias=time_spine_data_set_alias,
                joins_descs=(join_desc,),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_join_to_base_output_node(self, node: JoinToBaseOutputNode) -> SqlDataSet:
        """Generates the query that realizes the behavior of the JoinToStandardOutputNode."""
        # Keep a mapping between the table aliases that would be used in the query and the MDO instances in that source.
        # e.g. when building "FROM from_table a JOIN right_table b", the value for key "a" would be the instances in
        # "from_table"
        table_alias_to_instance_set: OrderedDict[str, InstanceSet] = OrderedDict()

        # Convert the dataflow from the left node to a DataSet and add context for it to table_alias_to_instance_set
        # A DataSet is a bundle of the SQL query (in object form) and the MDO instances that the SQL query contains.
        from_data_set = node.left_node.accept(self)
        from_data_set_alias = self._next_unique_table_alias()
        table_alias_to_instance_set[from_data_set_alias] = from_data_set.instance_set

        # Build the join descriptions for the SqlQueryPlan - different from node.join_descriptions which are the join
        # descriptions from the dataflow plan.
        sql_join_descs: List[SqlJoinDescription] = []

        # The dataflow plan describes how the data sets coming from the parent nodes should be joined together. Use
        # those descriptions to convert them to join descriptions for the SQL query plan.
        for join_description in node.join_targets:
            join_on_entity = join_description.join_on_entity

            right_node_to_join: BaseOutput = join_description.join_node
            right_data_set: SqlDataSet = right_node_to_join.accept(self)
            right_data_set_alias = self._next_unique_table_alias()

            sql_join_descs.append(
                SqlQueryPlanJoinBuilder.make_base_output_join_description(
                    left_data_set=AnnotatedSqlDataSet(data_set=from_data_set, alias=from_data_set_alias),
                    right_data_set=AnnotatedSqlDataSet(data_set=right_data_set, alias=right_data_set_alias),
                    join_description=join_description,
                )
            )

            # Remove the linkable instances with the join_on_entity as the leading link as the next step adds the
            # link. This is to avoid cases where there is a primary entity and a dimension in the data set, and we
            # create an instance in the next step that has the same entity link.
            # e.g. a data set has the dimension "listing__country_latest" and "listing" is a primary entity in the
            # data set. The next step would create an instance like "listing__listing__country_latest" without this
            # filter.

            # logger.error(f"before filter is:\n{pformat_big_objects(right_data_set.instance_set.spec_set)}")
            right_data_set_instance_set_filtered = FilterLinkableInstancesWithLeadingLink(
                entity_link=join_on_entity,
            ).transform(right_data_set.instance_set)
            # logger.error(f"after filter is:\n{pformat_big_objects(right_data_set_instance_set_filtered.spec_set)}")

            # After the right data set is joined to the "from" data set, we need to change the links for some of the
            # instances that represent the right data set. For example, if the "from" data set contains the "bookings"
            # measure instance and the right dataset contains the "country" dimension instance, then after the join,
            # the output data set should have the "country" dimension instance with the "user_id" entity link
            # (if "user_id" equality was the join condition). "country" -> "user_id__country"
            right_data_set_instance_set_after_join = right_data_set_instance_set_filtered.transform(
                AddLinkToLinkableElements(join_on_entity=join_on_entity)
            )
            table_alias_to_instance_set[right_data_set_alias] = right_data_set_instance_set_after_join

        from_data_set_output_instance_set = from_data_set.instance_set.transform(
            FilterElements(include_specs=from_data_set.instance_set.spec_set)
        )

        # Change the aggregation state for the measures to be partially aggregated if it was previously aggregated
        # since we removed the entities and added the dimensions. The dimensions could have the same value for
        # multiple rows, so we'll need to re-aggregate.
        from_data_set_output_instance_set = from_data_set_output_instance_set.transform(
            ChangeMeasureAggregationState(
                {
                    AggregationState.NON_AGGREGATED: AggregationState.NON_AGGREGATED,
                    AggregationState.COMPLETE: AggregationState.PARTIAL,
                    AggregationState.PARTIAL: AggregationState.PARTIAL,
                }
            )
        )

        table_alias_to_instance_set[from_data_set_alias] = from_data_set_output_instance_set

        # Construct the data set that contains the updated instances and the SQL nodes that should go in the various
        # clauses.
        return SqlDataSet(
            instance_set=InstanceSet.merge(list(table_alias_to_instance_set.values())),
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=create_select_columns_for_instance_sets(
                    self._column_association_resolver, table_alias_to_instance_set
                ),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=tuple(sql_join_descs),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_join_aggregated_measures_by_groupby_columns_node(  # noqa: D
        self, node: JoinAggregatedMeasuresByGroupByColumnsNode
    ) -> SqlDataSet:
        """Generates the query that realizes the behavior of the JoinAggregatedMeasuresByGroupByColumnsNode.

        This node is a straight inner join against all of the columns used for grouping in the input
        aggregation steps. Every column should be used, and at this point all inputs are fully aggregated,
        meaning we can make assumptions about things like NULL handling and there being one row per value
        set in each semantic model.

        In addition, this is used in cases where we expect a final metric to be computed using these
        measures as input. Therefore, we make the assumption that any mismatch should be discarded, as
        the behavior of the metric will be undefined in that case. This is why the INNER JOIN type is
        appropriate - if a dimension value set exists in one aggregated set but not the other, there is
        no sensible metric value for that dimension set.
        """
        assert len(node.parent_nodes) > 1, "This cannot happen, the node initializer would have failed"

        table_alias_to_instance_set: OrderedDict[str, InstanceSet] = OrderedDict()

        from_data_set: SqlDataSet = node.parent_nodes[0].accept(self)
        from_data_set_alias = self._next_unique_table_alias()
        table_alias_to_instance_set[from_data_set_alias] = from_data_set.instance_set
        join_aliases = [column.column_name for column in from_data_set.groupable_column_associations]
        use_cross_join = len(join_aliases) == 0

        sql_join_descs: List[SqlJoinDescription] = []
        for aggregated_node in node.parent_nodes[1:]:
            right_data_set: SqlDataSet = aggregated_node.accept(self)
            right_data_set_alias = self._next_unique_table_alias()
            right_column_names = {column.column_name for column in right_data_set.groupable_column_associations}
            if right_column_names != set(join_aliases):
                # TODO test multi-hop dimensions and address any issues. For now, let's raise an exception.
                raise ValueError(
                    f"We only support joins where all dimensions have the same name, but we got {right_column_names} "
                    f"and {join_aliases}, which differ by {right_column_names.difference(set(join_aliases))}!"
                )
            # sort column names to ensure consistent join ordering for ease of debugging and testing
            ordered_right_column_names = sorted(right_column_names)
            column_equality_descriptions = [
                ColumnEqualityDescription(
                    left_column_alias=colname, right_column_alias=colname, treat_nulls_as_equal=True
                )
                for colname in ordered_right_column_names
            ]
            sql_join_descs.append(
                SqlQueryPlanJoinBuilder.make_column_equality_sql_join_description(
                    right_source_node=right_data_set.sql_select_node,
                    right_source_alias=right_data_set_alias,
                    left_source_alias=from_data_set_alias,
                    column_equality_descriptions=column_equality_descriptions,
                    join_type=SqlJoinType.INNER if not use_cross_join else SqlJoinType.CROSS_JOIN,
                )
            )
            # All groupby columns are shared by all inputs, so we only want the measure/metric columns
            # from the semantic models on the right side of the join
            table_alias_to_instance_set[right_data_set_alias] = InstanceSet(
                measure_instances=right_data_set.instance_set.measure_instances,
                metric_instances=right_data_set.instance_set.metric_instances,
            )

        return SqlDataSet(
            instance_set=InstanceSet.merge(list(table_alias_to_instance_set.values())),
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=create_select_columns_for_instance_sets(
                    self._column_association_resolver, table_alias_to_instance_set
                ),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=tuple(sql_join_descs),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_aggregate_measures_node(self, node: AggregateMeasuresNode) -> SqlDataSet:
        """Generates the query that realizes the behavior of AggregateMeasuresNode.

        This will produce a query that aggregates all measures from a given input semantic model per the
        measure spec

        In the event the input aggregations are applied to measures with aliases set, in case of, e.g.,
        a constraint applied to one instance of the measure but not another one, this method will
        apply the rename in the select statement for this node, and propagate that further along via an
        instance set transform to rename the measures.

        Any node operating on the output of this node will need to use the measure aliases instead of
        the measure names as references.

        """
        # Get the data from the parent, and change measure instances to the aggregated state.
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        aggregated_instance_set = from_data_set.instance_set.transform(
            ChangeMeasureAggregationState(
                {
                    AggregationState.NON_AGGREGATED: AggregationState.COMPLETE,
                    AggregationState.COMPLETE: AggregationState.COMPLETE,
                    AggregationState.PARTIAL: AggregationState.COMPLETE,
                }
            )
        )
        # Also, the columns should always follow the resolver format.
        aggregated_instance_set = aggregated_instance_set.transform(
            ChangeAssociatedColumns(self._column_association_resolver)
        )

        from_data_set_alias = self._next_unique_table_alias()

        # Convert the instance set into a set of select column statements with updated aliases
        # Note any measure with an alias requirement will be recast at this point, and
        # downstream consumers of the resulting node must therefore request aggregated measures
        # by their appropriate aliases
        select_column_set: SelectColumnSet = aggregated_instance_set.transform(
            CreateSelectColumnsWithMeasuresAggregated(
                table_alias=from_data_set_alias,
                column_resolver=self._column_association_resolver,
                semantic_model_lookup=self._semantic_model_lookup,
                metric_input_measure_specs=node.metric_input_measure_specs,
            )
        )

        if any((spec.alias for spec in node.metric_input_measure_specs)):
            # This is a little silly, but we need to update the column instance set with the new aliases
            # There are a number of refactoring options - simplest is to consolidate this with
            # ChangeMeasureAggregationState, assuming there are no ordering dependencies up above
            aggregated_instance_set = aggregated_instance_set.transform(
                AliasAggregatedMeasures(metric_input_measure_specs=node.metric_input_measure_specs)
            )
            # and make sure we follow the resolver format for any newly aliased measures....
            aggregated_instance_set = aggregated_instance_set.transform(
                ChangeAssociatedColumns(self._column_association_resolver)
            )

        return SqlDataSet(
            instance_set=aggregated_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This will generate expressions with the appropriate aggregation functions e.g. SUM()
                select_columns=select_column_set.as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                # This will generate expressions to group by the columns that don't correspond to a measure instance.
                group_bys=select_column_set.without_measure_columns().as_tuple(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> SqlDataSet:
        """Generates the query that realizes the behavior of ComputeMetricsNode."""
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        from_data_set_alias = self._next_unique_table_alias()

        # TODO: Check that all measures for the metrics are in the input instance set
        # The desired output instance set has no measures, so create a copy with those removed.
        output_instance_set: InstanceSet = from_data_set.instance_set.transform(RemoveMeasures())

        # Also, the output columns should always follow the resolver format.
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))
        output_instance_set = output_instance_set.transform(RemoveMetrics())

        non_metric_select_column_set: SelectColumnSet = output_instance_set.transform(
            CreateSelectColumnsForInstances(
                table_alias=from_data_set_alias,
                column_resolver=self._column_association_resolver,
            )
        )

        # Add select columns that would compute the metrics to the select columns.
        metric_select_columns = []
        metric_instances = []
        for metric_spec in node.metric_specs:
            metric = self._metric_lookup.get_metric(metric_spec.as_reference)

            metric_expr: Optional[SqlExpressionNode] = None
            if metric.type is MetricType.RATIO:
                numerator = metric.type_params.numerator
                denominator = metric.type_params.denominator
                assert (
                    numerator is not None and denominator is not None
                ), "Missing numerator or denominator for ratio metric, this should have been caught in validation!"
                numerator_column_name = self._column_association_resolver.resolve_spec(
                    MetricSpec.from_reference(numerator.post_aggregation_reference)
                ).column_name
                denominator_column_name = self._column_association_resolver.resolve_spec(
                    MetricSpec.from_reference(denominator.post_aggregation_reference)
                ).column_name

                metric_expr = SqlRatioComputationExpression(
                    numerator=SqlColumnReferenceExpression(
                        SqlColumnReference(
                            table_alias=from_data_set_alias,
                            column_name=numerator_column_name,
                        )
                    ),
                    denominator=SqlColumnReferenceExpression(
                        SqlColumnReference(
                            table_alias=from_data_set_alias,
                            column_name=denominator_column_name,
                        )
                    ),
                )
            elif metric.type is MetricType.SIMPLE:
                if len(metric.input_measures) > 0:
                    assert (
                        len(metric.input_measures) == 1
                    ), "Measure proxy metrics should always source from exactly 1 measure."
                    expr = self._column_association_resolver.resolve_spec(
                        MeasureSpec(
                            element_name=metric.input_measures[0].post_aggregation_measure_reference.element_name
                        )
                    ).column_name
                else:
                    expr = metric.name
                # Use a column reference to improve query optimization.
                metric_expr = SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=from_data_set_alias,
                        column_name=expr,
                    )
                )
            elif metric.type is MetricType.CUMULATIVE:
                assert (
                    len(metric.measure_references) == 1
                ), "Cumulative metrics should always source from exactly 1 measure."
                expr = self._column_association_resolver.resolve_spec(
                    MeasureSpec(element_name=metric.input_measures[0].post_aggregation_measure_reference.element_name)
                ).column_name
                metric_expr = SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=from_data_set_alias,
                        column_name=expr,
                    )
                )
            elif metric.type is MetricType.DERIVED:
                assert metric.type_params.expr
                metric_expr = SqlStringExpression(sql_expr=metric.type_params.expr)
            else:
                assert_values_exhausted(metric.type)

            assert metric_expr

            output_column_association = self._column_association_resolver.resolve_spec(metric_spec.alias_spec)
            metric_select_columns.append(
                SqlSelectColumn(
                    expr=metric_expr,
                    column_alias=output_column_association.column_name,
                )
            )
            metric_instances.append(
                MetricInstance(
                    associated_columns=(output_column_association,),
                    defined_from=(MetricModelReference(metric_name=metric_spec.element_name),),
                    spec=metric_spec.alias_spec,
                )
            )
        output_instance_set = output_instance_set.transform(AddMetrics(metric_instances))

        combined_select_column_set = non_metric_select_column_set.merge(
            SelectColumnSet(metric_columns=metric_select_columns)
        )

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=combined_select_column_set.as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> SqlDataSet:  # noqa: D
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        output_instance_set = from_data_set.instance_set
        from_data_set_alias = self._next_unique_table_alias()

        # Also, the output columns should always follow the resolver format.
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        order_by_descriptions = []
        for order_by_spec in node.order_by_specs:
            order_by_descriptions.append(
                SqlOrderByDescription(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias=from_data_set_alias,
                            column_name=self._column_association_resolver.resolve_spec(
                                order_by_spec.instance_spec
                            ).column_name,
                        )
                    ),
                    desc=order_by_spec.descending,
                )
            )

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=output_instance_set.transform(
                    CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
                ).as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=None,
                order_bys=tuple(order_by_descriptions),
                limit=node.limit,
            ),
        )

    def visit_write_to_result_dataframe_node(self, node: WriteToResultDataframeNode) -> SqlDataSet:
        """This is an operation that can't be represented as an SQL query.

        Instead, it should be handled in the execution plan as an operation that runs an SQL query and saves it to
        a dataframe.
        """
        raise RuntimeError("This node type is not supported.")

    def visit_write_to_result_table_node(self, node: WriteToResultTableNode) -> SqlDataSet:
        """Similar to visit_write_to_result_dataframe_node()."""
        raise RuntimeError("This node type is not supported.")

    def visit_pass_elements_filter_node(self, node: FilterElementsNode) -> SqlDataSet:
        """Generates the query that realizes the behavior of FilterElementsNode."""
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        output_instance_set = from_data_set.instance_set.transform(FilterElements(node.include_specs))
        from_data_set_alias = self._next_unique_table_alias()

        # Also, the output columns should always follow the resolver format.
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        # This creates select expressions for all columns referenced in the instance set.
        select_columns = output_instance_set.transform(
            CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
        ).as_tuple()

        # If distinct values requested, group by all select columns.
        group_bys = select_columns if node.distinct else ()
        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=select_columns,
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=group_bys,
                where=None,
                order_bys=(),
            ),
        )

    def visit_where_constraint_node(self, node: WhereConstraintNode) -> SqlDataSet:
        """Adds where clause to SQL statement from parent node."""
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        output_instance_set = from_data_set.instance_set
        from_data_set_alias = self._next_unique_table_alias()

        column_associations_in_where_sql: Sequence[ColumnAssociation] = CreateColumnAssociations(
            column_association_resolver=self._column_association_resolver
        ).transform(spec_set=node.where.linkable_spec_set.as_spec_set)

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=output_instance_set.transform(
                    CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
                ).as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=SqlStringExpression(
                    sql_expr=node.where.where_sql,
                    used_columns=tuple(
                        column_association.column_name for column_association in column_associations_in_where_sql
                    ),
                    bind_parameters=node.where.bind_parameters,
                ),
                order_bys=(),
            ),
        )

    def _make_select_columns_for_metrics(
        self,
        table_alias_to_metric_specs: OrderedDict[str, Sequence[MetricSpec]],
        aggregation_type: Optional[AggregationType],
    ) -> List[SqlSelectColumn]:
        """Creates select columns that get the given metric using the given table alias.

        e.g.

        with table_alias_to_metric_specs = {"a": MetricSpec(element_name="bookings")}

        ->

        a.bookings AS bookings
        """
        select_columns = []
        for table_alias, metric_specs in table_alias_to_metric_specs.items():
            for metric_spec in metric_specs:
                metric_column_name = self._column_association_resolver.resolve_spec(metric_spec).column_name
                column_reference_expression = SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(
                        table_alias=table_alias,
                        column_name=metric_column_name,
                    )
                )
                if aggregation_type:
                    select_expression: SqlExpressionNode = SqlFunctionExpression.build_expression_from_aggregation_type(
                        aggregation_type=aggregation_type, sql_column_expression=column_reference_expression
                    )
                else:
                    select_expression = column_reference_expression

                select_columns.append(
                    SqlSelectColumn(
                        expr=select_expression,
                        column_alias=metric_column_name,
                    )
                )
        return select_columns

    def visit_combine_metrics_node(self, node: CombineMetricsNode) -> SqlDataSet:
        """Join computed metric datasets together to return a single dataset containing all metrics.

        This node may exist in one of two situations: when metrics need to be combined in order to produce a single
        dataset with all required inputs for a derived metric (in which case the join type is INNER), or when
        metrics need to be combined in order to produce a single dataset of output for downstream consumption by
        the end user, in which case we will use FULL OUTER JOIN.

        In the case of a multi-data-source FULL OUTER JOIN the join key will be a coalesced set of all previously
        seen dimension values. For example:
            FROM (
              ...
            ) subq_9
            FULL OUTER JOIN (
              ...
            ) subq_10
            ON
              subq_9.is_instant = subq_10.is_instant
              AND subq_9.ds = subq_10.ds
            FULL OUTER JOIN (
              ...
            ) subq_11
            ON
              COALESCE(subq_9.is_instant, subq_10.is_instant) = subq_11.is_instant
              AND COALESCE(subq_9.ds, subq_10.ds) = subq_11.ds

        Whenever these nodes are joined using a FULL OUTER JOIN, we must also do a subsequent re-aggregation pass to
        deduplicate the dimension value outputs across different metrics. This can happen if one or more of the
        dimensions contains a NULL value. In that case, the FULL OUTER JOIN condition will fail, because NULL = NULL
        returns NULL. Unfortunately, there's no way to do a robust NULL-safe comparison across engines in a FULL
        OUTER JOIN context, because many engines do not support complex ON conditions or other techniques we might
        use to apply a sentinel value for NULL to NULL comparisons.
        """
        assert (
            len(node.parent_nodes) > 1
        ), "Shouldn't have a CombineMetricsNode in the dataflow plan if there's only 1 parent."

        parent_data_sets: List[AnnotatedSqlDataSet] = []
        table_alias_to_metric_specs: OrderedDict[str, Sequence[MetricSpec]] = OrderedDict()

        for parent_node in node.parent_nodes:
            parent_sql_data_set = parent_node.accept(self)
            table_alias = self._next_unique_table_alias()
            parent_data_sets.append(AnnotatedSqlDataSet(data_set=parent_sql_data_set, alias=table_alias))
            table_alias_to_metric_specs[table_alias] = parent_sql_data_set.instance_set.spec_set.metric_specs

        # When we create the components of the join that combines metrics it will be one of INNER, FULL OUTER,
        # or CROSS JOIN. Order doesn't matter for these join types, so we will use the first element in the FROM
        # clause and create join descriptions from the rest.
        from_data_set = parent_data_sets[0]
        join_data_sets = parent_data_sets[1:]

        # Sanity check that all parents have the same linkable specs before building the join descriptions.
        linkable_specs = from_data_set.data_set.instance_set.spec_set.linkable_specs
        assert all(
            [set(x.data_set.instance_set.spec_set.linkable_specs) == set(linkable_specs) for x in join_data_sets]
        ), "All parent nodes should have the same set of linkable instances since all values are coalesced."

        linkable_spec_set = from_data_set.data_set.instance_set.spec_set.transform(SelectOnlyLinkableSpecs())
        join_type = SqlJoinType.CROSS_JOIN if len(linkable_spec_set.all_specs) == 0 else node.join_type

        joins_descriptions: List[SqlJoinDescription] = []
        # TODO: refactor this loop into SqlQueryPlanJoinBuilder
        column_associations = tuple(
            self._column_association_resolver.resolve_spec(spec) for spec in linkable_spec_set.all_specs
        )
        column_names = tuple(association.column_name for association in column_associations)
        aliases_seen = [from_data_set.alias]
        for join_data_set in join_data_sets:
            joins_descriptions.append(
                SqlQueryPlanJoinBuilder.make_combine_metrics_join_description(
                    from_data_set=from_data_set,
                    join_data_set=join_data_set,
                    join_type=join_type,
                    column_names=column_names,
                    table_aliases_for_coalesce=aliases_seen,
                )
            )
            aliases_seen.append(join_data_set.alias)

        # We can merge all parent instances since the common linkable instances will be de-duped.
        output_instance_set = InstanceSet.merge([x.data_set.instance_set for x in parent_data_sets])
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        metric_aggregation_type = AggregationType.MAX if node.join_type is SqlJoinType.FULL_OUTER else None
        metric_select_column_set = SelectColumnSet(
            metric_columns=self._make_select_columns_for_metrics(
                table_alias_to_metric_specs, aggregation_type=metric_aggregation_type
            )
        )
        linkable_select_column_set = linkable_spec_set.transform(
            CreateSelectCoalescedColumnsForLinkableSpecs(
                column_association_resolver=self._column_association_resolver,
                table_aliases=[x.alias for x in parent_data_sets],
            )
        )
        combined_select_column_set = linkable_select_column_set.merge(metric_select_column_set)

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=combined_select_column_set.as_tuple(),
                from_source=from_data_set.data_set.sql_select_node,
                from_source_alias=from_data_set.alias,
                joins_descs=tuple(joins_descriptions),
                group_bys=linkable_select_column_set.as_tuple() if node.join_type is SqlJoinType.FULL_OUTER else (),
                where=None,
                order_bys=(),
            ),
        )

    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode) -> SqlDataSet:
        """Convert ConstrainTimeRangeNode to a SqlDataSet by building the time constraint comparison.

        Use the smallest time granularity to build the comparison since that's what was used in the semantic model
        definition and it wouldn't have a DATE_TRUNC() in the expression. We want to build this:

            ds >= '2020-01-01' AND ds <= '2020-02-01'

        instead of this: DATE_TRUNC('month', ds) >= '2020-01-01' AND DATE_TRUNC('month', ds <= '2020-02-01')
        """
        from_data_set: SqlDataSet = node.parent_node.accept(self)
        from_data_set_alias = self._next_unique_table_alias()

        time_dimension_instances_for_metric_time = sorted(
            from_data_set.metric_time_dimension_instances,
            key=lambda x: x.spec.time_granularity.to_int(),
        )

        assert (
            len(time_dimension_instances_for_metric_time) > 0
        ), "No metric time dimensions found in the input data set for this node"

        time_dimension_instance_for_metric_time = time_dimension_instances_for_metric_time[0]

        # Build an expression like "ds >= CAST('2020-01-01' AS TIMESTAMP) AND ds <= CAST('2020-01-02' AS TIMESTAMP)"
        constrain_metric_time_column_condition = _make_time_range_comparison_expr(
            table_alias=from_data_set_alias,
            column_alias=time_dimension_instance_for_metric_time.associated_column.column_name,
            time_range_constraint=node.time_range_constraint,
        )

        output_instance_set = from_data_set.instance_set
        # Output columns should always follow the resolver format.
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=output_instance_set.transform(
                    CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
                ).as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=constrain_metric_time_column_condition,
                order_bys=(),
            ),
        )

    def convert_to_sql_query_plan(
        self,
        sql_engine_type: SqlEngine,
        sql_query_plan_id: str,
        dataflow_plan_node: Union[BaseOutput, ComputedMetricsOutput],
        optimization_level: SqlQueryOptimizationLevel = SqlQueryOptimizationLevel.O4,
    ) -> SqlQueryPlan:
        """Create an SQL query plan that represents the computation up to the given dataflow plan node."""
        sql_select_node: SqlQueryPlanNode = dataflow_plan_node.accept(self).sql_select_node

        # TODO: Make this a more generally accessible attribute instead of checking against the
        # BigQuery-ness of the engine
        use_column_alias_in_group_by = sql_engine_type is SqlEngine.BIGQUERY

        for optimizer in SqlQueryOptimizerConfiguration.optimizers_for_level(
            optimization_level, use_column_alias_in_group_by=use_column_alias_in_group_by
        ):
            logger.info(f"Applying optimizer: {optimizer.__class__.__name__}")
            sql_select_node = optimizer.optimize(sql_select_node)

        return SqlQueryPlan(plan_id=sql_query_plan_id, render_node=sql_select_node)

    def visit_metric_time_dimension_transform_node(self, node: MetricTimeDimensionTransformNode) -> SqlDataSet:
        """Implement the behavior of the MetricTimeDimensionTransformNode.

        This node will create an output data set that is similar to the input data set, but the measure instances it
        contains is a subset of the input data set. Only measure instances that have an aggregation time dimension
        matching the one defined in the node will be passed. In addition, an additional time dimension instance for
        "metric time" will be included. See DataSet.metric_time_dimension_reference().
        """
        input_data_set: SqlDataSet = node.parent_node.accept(self)

        # Find which measures have an aggregation time dimension that is the same as the one specified in the node.
        # Only these measures will be in the output data set.
        output_measure_instances = []
        for measure_instance in input_data_set.instance_set.measure_instances:
            semantic_model = self._semantic_model_lookup.get_by_reference(
                semantic_model_reference=measure_instance.origin_semantic_model_reference.semantic_model_reference
            )
            assert semantic_model is not None, (
                f"{measure_instance} was defined from {measure_instance.origin_semantic_model_reference}, but that"
                f"can't be found"
            )
            aggregation_time_dimension_for_measure = semantic_model.checked_agg_time_dimension_for_measure(
                measure_reference=measure_instance.spec.as_reference
            )
            if aggregation_time_dimension_for_measure == node.aggregation_time_dimension_reference:
                output_measure_instances.append(measure_instance)

        if len(output_measure_instances) == 0:
            raise RuntimeError(
                f"No measure instances in the input source match the aggregation time dimension "
                f"{node.aggregation_time_dimension_reference}. Check if the dataflow plan was constructed correctly."
            )

        # Find time dimension instances that refer to the same dimension as the one specified in the node.
        matching_time_dimension_instances = []
        for time_dimension_instance in input_data_set.instance_set.time_dimension_instances:
            # The specification for the time dimension to use for aggregation is the local one.
            if (
                len(time_dimension_instance.spec.entity_links) == 0
                and time_dimension_instance.spec.reference == node.aggregation_time_dimension_reference
            ):
                matching_time_dimension_instances.append(time_dimension_instance)

        output_time_dimension_instances: List[TimeDimensionInstance] = []
        output_time_dimension_instances.extend(input_data_set.instance_set.time_dimension_instances)
        output_column_to_input_column: OrderedDict[str, str] = OrderedDict()

        # For those matching time dimension instances, create the analog metric time dimension instances for the output.
        for matching_time_dimension_instance in matching_time_dimension_instances:
            metric_time_dimension_spec = DataSet.metric_time_dimension_spec(
                time_granularity=matching_time_dimension_instance.spec.time_granularity,
                date_part=matching_time_dimension_instance.spec.date_part,
            )
            metric_time_dimension_column_association = self._column_association_resolver.resolve_spec(
                metric_time_dimension_spec
            )
            output_time_dimension_instances.append(
                TimeDimensionInstance(
                    defined_from=matching_time_dimension_instance.defined_from,
                    associated_columns=(self._column_association_resolver.resolve_spec(metric_time_dimension_spec),),
                    spec=metric_time_dimension_spec,
                )
            )
            output_column_to_input_column[
                metric_time_dimension_column_association.column_name
            ] = matching_time_dimension_instance.associated_column.column_name

        output_instance_set = InstanceSet(
            measure_instances=tuple(output_measure_instances),
            dimension_instances=input_data_set.instance_set.dimension_instances,
            time_dimension_instances=tuple(output_time_dimension_instances),
            entity_instances=input_data_set.instance_set.entity_instances,
            metric_instances=input_data_set.instance_set.metric_instances,
        )
        output_instance_set = ChangeAssociatedColumns(self._column_association_resolver).transform(output_instance_set)

        from_data_set_alias = self._next_unique_table_alias()

        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                # This creates select expressions for all columns referenced in the instance set.
                select_columns=CreateSelectColumnsForInstances(
                    column_resolver=self._column_association_resolver,
                    table_alias=from_data_set_alias,
                    output_to_input_column_mapping=output_column_to_input_column,
                )
                .transform(output_instance_set)
                .as_tuple(),
                from_source=input_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_semi_additive_join_node(self, node: SemiAdditiveJoinNode) -> SqlDataSet:
        """Implements the behaviour of SemiAdditiveJoinNode.

        This node will get the build a data set row filtered by the aggregate function on the
        specified dimension that is non-additive. Then that dataset would be joined with the input data
        on that dimension along with grouping by entities that are also passed in.
        """
        from_data_set: SqlDataSet = node.parent_node.accept(self)

        from_data_set_alias = self._next_unique_table_alias()

        # Get the output_instance_set of the parent_node
        output_instance_set = from_data_set.instance_set
        output_instance_set = output_instance_set.transform(ChangeAssociatedColumns(self._column_association_resolver))

        # Build the JoinDescriptions to handle the row base filtering on the output_data_set
        inner_join_data_set_alias = self._next_unique_table_alias()

        column_equality_descriptions: List[ColumnEqualityDescription] = []

        # Build Time Dimension SqlSelectColumn
        time_dimension_column_name = self.column_association_resolver.resolve_spec(node.time_dimension_spec).column_name
        join_time_dimension_column_name = self.column_association_resolver.resolve_spec(
            node.time_dimension_spec.with_aggregation_state(AggregationState.COMPLETE),
        ).column_name
        time_dimension_select_column = SqlSelectColumn(
            expr=SqlFunctionExpression.build_expression_from_aggregation_type(
                aggregation_type=node.agg_by_function,
                sql_column_expression=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=inner_join_data_set_alias,
                        column_name=time_dimension_column_name,
                    ),
                ),
            ),
            column_alias=join_time_dimension_column_name,
        )
        column_equality_descriptions.append(
            ColumnEqualityDescription(
                left_column_alias=time_dimension_column_name,
                right_column_alias=join_time_dimension_column_name,
            )
        )

        # Build optional window grouping SqlSelectColumn
        entity_select_columns: List[SqlSelectColumn] = []
        for entity_spec in node.entity_specs:
            entity_column_name = self.column_association_resolver.resolve_spec(entity_spec).column_name
            entity_select_columns.append(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        SqlColumnReference(
                            table_alias=inner_join_data_set_alias,
                            column_name=entity_column_name,
                        ),
                    ),
                    column_alias=entity_column_name,
                )
            )
            column_equality_descriptions.append(
                ColumnEqualityDescription(
                    left_column_alias=entity_column_name,
                    right_column_alias=entity_column_name,
                )
            )

        # Propogate additional group by during query time of the non-additive time dimension
        queried_time_dimension_select_column: Optional[SqlSelectColumn] = None
        if node.queried_time_dimension_spec:
            query_time_dimension_column_name = self.column_association_resolver.resolve_spec(
                node.queried_time_dimension_spec
            ).column_name
            queried_time_dimension_select_column = SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=inner_join_data_set_alias,
                        column_name=query_time_dimension_column_name,
                    ),
                ),
                column_alias=query_time_dimension_column_name,
            )

        row_filter_group_bys = tuple(entity_select_columns)
        if queried_time_dimension_select_column:
            row_filter_group_bys += (queried_time_dimension_select_column,)
        # Construct SelectNode for Row filtering
        row_filter_sql_select_node = SqlSelectStatementNode(
            description=f"Filter row on {node.agg_by_function.name}({time_dimension_column_name})",
            select_columns=row_filter_group_bys + (time_dimension_select_column,),
            from_source=from_data_set.sql_select_node,
            from_source_alias=inner_join_data_set_alias,
            joins_descs=(),
            group_bys=row_filter_group_bys,
            where=None,
            order_bys=(),
        )

        join_data_set_alias = self._next_unique_table_alias()
        sql_join_desc = SqlQueryPlanJoinBuilder.make_column_equality_sql_join_description(
            right_source_node=row_filter_sql_select_node,
            left_source_alias=from_data_set_alias,
            right_source_alias=join_data_set_alias,
            column_equality_descriptions=column_equality_descriptions,
            join_type=SqlJoinType.INNER,
        )
        return SqlDataSet(
            instance_set=output_instance_set,
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=output_instance_set.transform(
                    CreateSelectColumnsForInstances(from_data_set_alias, self._column_association_resolver)
                ).as_tuple(),
                from_source=from_data_set.sql_select_node,
                from_source_alias=from_data_set_alias,
                joins_descs=(sql_join_desc,),
                group_bys=(),
                where=None,
                order_bys=(),
            ),
        )

    def visit_join_to_time_spine_node(self, node: JoinToTimeSpineNode) -> SqlDataSet:  # noqa: D
        parent_data_set = node.parent_node.accept(self)
        parent_alias = self._next_unique_table_alias()

        # Build time spine dataset
        metric_time_dimension_instance: Optional[TimeDimensionInstance] = None
        for instance in parent_data_set.metric_time_dimension_instances:
            if len(instance.spec.entity_links) == 0:
                # Use the instance with the lowest granularity
                if not metric_time_dimension_instance or (
                    instance.spec.time_granularity < metric_time_dimension_instance.spec.time_granularity
                ):
                    metric_time_dimension_instance = instance
        assert (
            metric_time_dimension_instance
        ), "Can't query offset metric without a time dimension. Validations should have prevented this."
        metric_time_dimension_column_name = self.column_association_resolver.resolve_spec(
            metric_time_dimension_instance.spec
        ).column_name
        time_spine_alias = self._next_unique_table_alias()
        time_spine_dataset = self._make_time_spine_data_set(
            metric_time_dimension_instance=metric_time_dimension_instance,
            metric_time_dimension_column_name=metric_time_dimension_column_name,
            time_spine_source=self._time_spine_source,
            time_range_constraint=node.time_range_constraint,
        )

        # Build join expression
        join_description = SqlQueryPlanJoinBuilder.make_join_to_time_spine_join_description(
            node=node,
            time_spine_alias=time_spine_alias,
            metric_time_dimension_column_name=metric_time_dimension_column_name,
            parent_sql_select_node=parent_data_set.sql_select_node,
            parent_alias=parent_alias,
        )

        # Use all instances EXCEPT metric_time from parent data set.
        non_metric_time_parent_instance_set = InstanceSet(
            measure_instances=parent_data_set.instance_set.measure_instances,
            dimension_instances=parent_data_set.instance_set.dimension_instances,
            time_dimension_instances=tuple(
                time_dimension_instance
                for time_dimension_instance in parent_data_set.instance_set.time_dimension_instances
                if time_dimension_instance.spec.element_name != DataSet.metric_time_dimension_reference().element_name
            ),
            entity_instances=parent_data_set.instance_set.entity_instances,
            metric_instances=parent_data_set.instance_set.metric_instances,
            metadata_instances=parent_data_set.instance_set.metadata_instances,
        )
        non_metric_time_select_columns = create_select_columns_for_instance_sets(
            self._column_association_resolver, OrderedDict({parent_alias: non_metric_time_parent_instance_set})
        )

        # Use metric_time column from time spine.
        assert (
            len(time_spine_dataset.instance_set.time_dimension_instances) == 1
            and len(time_spine_dataset.sql_select_node.select_columns) == 1
        ), "Time spine dataset not configured properly. Expected exactly one column."
        time_dim_instance = time_spine_dataset.instance_set.time_dimension_instances[0]
        time_spine_column_select_expr: Union[
            SqlColumnReferenceExpression, SqlDateTruncExpression
        ] = SqlColumnReferenceExpression(
            SqlColumnReference(table_alias=time_spine_alias, column_name=time_dim_instance.spec.qualified_name)
        )

        # Add requested granularities (skip for default granularity) and date_parts.
        metric_time_select_columns = []
        metric_time_dimension_instances = []
        where: Optional[SqlExpressionNode] = None
        for metric_time_dimension_spec in node.metric_time_dimension_specs:
            # Apply granularity to SQL.
            if metric_time_dimension_spec.time_granularity == self._time_spine_source.time_column_granularity:
                select_expr: SqlExpressionNode = time_spine_column_select_expr
            else:
                select_expr = SqlDateTruncExpression(
                    time_granularity=metric_time_dimension_spec.time_granularity, arg=time_spine_column_select_expr
                )
                if node.offset_to_grain:
                    # Filter down to one row per granularity period
                    new_filter = SqlComparisonExpression(
                        left_expr=select_expr, comparison=SqlComparison.EQUALS, right_expr=time_spine_column_select_expr
                    )
                    if not where:
                        where = new_filter
                    else:
                        where = SqlLogicalExpression(operator=SqlLogicalOperator.OR, args=(where, new_filter))
            # Apply date_part to SQL.
            if metric_time_dimension_spec.date_part:
                select_expr = SqlExtractExpression(date_part=metric_time_dimension_spec.date_part, arg=select_expr)
            time_dim_spec = TimeDimensionSpec(
                element_name=time_dim_instance.spec.element_name,
                entity_links=time_dim_instance.spec.entity_links,
                time_granularity=metric_time_dimension_spec.time_granularity,
                date_part=metric_time_dimension_spec.date_part,
                aggregation_state=time_dim_instance.spec.aggregation_state,
            )
            time_dim_instance = TimeDimensionInstance(
                defined_from=time_dim_instance.defined_from,
                associated_columns=(self._column_association_resolver.resolve_spec(time_dim_spec),),
                spec=time_dim_spec,
            )
            metric_time_dimension_instances.append(time_dim_instance)
            metric_time_select_columns.append(
                SqlSelectColumn(expr=select_expr, column_alias=time_dim_instance.associated_column.column_name)
            )
        metric_time_instance_set = InstanceSet(time_dimension_instances=tuple(metric_time_dimension_instances))

        return SqlDataSet(
            instance_set=InstanceSet.merge([metric_time_instance_set, non_metric_time_parent_instance_set]),
            sql_select_node=SqlSelectStatementNode(
                description=node.description,
                select_columns=tuple(metric_time_select_columns) + non_metric_time_select_columns,
                from_source=time_spine_dataset.sql_select_node,
                from_source_alias=time_spine_alias,
                joins_descs=(join_description,),
                group_bys=(),
                order_bys=(),
                where=where,
            ),
        )



from __future__ import annotations

import pytest
from dbt_semantic_interfaces.dataclass_serialization import DataClassDeserializer, DataclassSerializer

from metricflow.sql.sql_bind_parameters import SqlBindParameter, SqlBindParameters, SqlBindParameterValue


@pytest.fixture
def serializer() -> DataclassSerializer:  # noqa: D
    return DataclassSerializer()


@pytest.fixture
def deserializer() -> DataClassDeserializer:  # noqa: D
    return DataClassDeserializer()


def test_serialization(  # noqa: D
    serializer: DataclassSerializer,
    deserializer: DataClassDeserializer,
) -> None:
    bind_parameters = SqlBindParameters(
        param_items=(
            SqlBindParameter("key0", SqlBindParameterValue.create_from_sql_column_type("value0")),
            SqlBindParameter("key1", SqlBindParameterValue.create_from_sql_column_type("value1")),
        )
    )
    serialized_obj = serializer.pydantic_serialize(bind_parameters)
    deserialized_obj = deserializer.pydantic_deserialize(
        dataclass_type=SqlBindParameters, serialized_obj=serialized_obj
    )
    assert bind_parameters == deserialized_obj
